#include "activity_estimator.h"

/*-----------------------------------------------------------------------------
 *  Copyright 2002 By Kara K.W. Poon and Steve Wilton
 *
 * Permission to use, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 * Permission is granted ONLY for non-commercial use.  For
 * commercial use, please contact the authors.
 *
 * This software may be distributed (but not offered for sale or transferred
 * for compensation) to third parties, provided such third parties agree to
 * abide by the terms and conditions of this notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHORS, AS
 * WELL AS THE UNIVERSITY OF BRITISH COLUMBIA AND THE
 * UNIVERSITY OF TORONTO DISCLAIM ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL THE
 * AUTHORS OR THE UNIVERSITY OF BRITISH COLUMBIA OR THE
 * UNIVERSITY OF TORONTO BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *---------------------------------------------------------------------------*/

/* 
	
	Project Topic: Flexible Power Model for Field Programmable Gate Arrays
	Institute:     The University of British Columbia, Canada
	Supervisor:    Steve J. E. Wilton

	Program Name:  probcalc.h and probcalc.cpp
	Function:      This is an activity generator for estimating the switching
                  activity at the output of each LUT. 

	Input file:    .blif file from SIS
	Output file:   .txt file (describes the probabilities and transition 
	                          densities of inputs and the ouput)
						.act file (records the probabilities and transition 
                             densities. This file will be used as an input
                             file for TV-pack for power estimation purpose) 			  
									  
	History:
	July, 2001     Program was first developed by Andy Yan

	February, 2002 Modified by Kara Poon to handle different LUT sizes and 
                  take the static probabiliy and transition density for 
						the primary inputs as input variables to this program.  

	April, 2002    Modified by Kara Poon to include the Filter mechanism
                  of the transition density model					
                
 */

/***************************************************************************/

/* Declare functions */

void build_table ();
/* 
	Function: Build a dynamic table based on the LUT size 
*/

void delete_table ();
/* 
	Function: Delete the dynamic table based on the LUT size 
*/

void built_LUT_Data (int numLut);
/* 
	Function:Allocate Memory for the LUT record
	Variable:
	numLut: Total number of LUTs in the circuit
*/

void delete_LUT_Data (int numLut);
/* 
	Function: Deallocate Memory for the LUT record
	Variable:
	numLut: Total number of LUTs in the circuit
*/

void int2str (int input_num,
				  int string_len,
				  char* outputstr);
/* 
	Function: Convert an integer to a character string
	Variables:
	input_num: input numerical number
	string_len: length of the binary string
	outputstr: output binary string for the input number
 */

void allocMem (int numLut, 
					int numLatch);
/* 
	Function: Allocate Memory for Luts and Latches
	Variables: 
	numLut: Number of LUTs
	numLatch: Number of MUXs
*/


void freeMem ();
/* 
	Function: Delete all dynamically allocated memory 	
*/

void addNet (char *buffer, 
				 double prob);
/* 
	Function: Add a signal name and its probability 
             value to the netList hash table
	Variables:
	buffer: signal name
	prob  : probability of the signal
*/

double getProb (char *signal);
/* 
	Function: Get the probability value based on the signal name
	Variables:
	signal: the signal name
*/
 
void addInName (char * buffer, 
					 int startpos, 
                int endpos, 
					 int inNamePtrPos,
                int lutIndex);
/* 
	Function: Add a input name to a lut
	Variables:
	buffer: an array of LUT input names
	startpos: starting position of the name in the array
	endpos: ending position of the name in the array
	inNamePtrPos: Pointer to the start of the input name
	lutIndex: index number represents the LUT
*/

void addOutName (char * buffer, 
					  int startpos, 
                 int endpos, 
					  int outNamePtrPos,
                 int lutIndex);
/* 
	Function: Add a output name to a lut
	Variables:
	buffer: an array of LUT output names
	startpos: starting position of the name in the array
	endpos: ending position of the name in the array
	outNamePtrPos: Pointer to the start of the output name
	lutIndex: index number represents the LUT
*/

void addLatch (char *buffer, 
					int startpos, 
					int endpos, 
					int ptrPos, 
               int index, 
					int in_out);

/* 
	Function: Add a input or output name to a latch
	Variables:
	buffer: an array of latch input or output names
	startpos: starting position of the name in the array
	endpos: ending position of the name in the array
	ptrPos: Pointer to the start of the input or output name
*/

void addInput (char *buffer, 
					int startpos, 
					int endpos);
/* 
	Function: Add input name to the netList hash table 
	Variables:
	buffer: an array of LUT input names
	startpos: starting position of the name in the array
	endpos: ending position of the name in the array
*/


void addHighEntry (char * buffer, 
						 int lutIndex);
/* 
	Function: Compute the "truth table" of the lut 
	Variables:
	buffer: character array that stored the function 
	        in .blif
	lutIndex: index number represents the LUT
*/


void orderLuts();
/* 
	Function: Connect the LUTs

*/

int calcProb (double err);
/* 
	Function: Calculate the probability that the output 
             is high for each LUT
	Variables:
	err; input error for feedback
*/

void calcProb_perLUT (int targetLut);


void calcChange();
/* 
	Function: Calculate the transition probability, 
	          which is 2p(p-1)

*/


void calcDensity();
/* 
	Function: Calculate the transition density x
	
*/

void calcDensity_perLUT (int targetLut);

double getDensity (char *signal);
/* 
	Function: Get the transition density value
	Variables:
	signal: the signal name
*/


void dumpNetList (char *outFile, 
						int outType);
/* 
	Function: dump values to file (*.act)
	Variables:
	outFile: output file name
	outType: 0 = output transition density only
	         1 = output probability only
				other (default) = output transition density 
				and probability 
				Note: VPR needs both transition density 
				      and probability values
*/

void dump2File (char *outFile);
/* 
	Function: Write log data to *.txt file 
	Variables:
	outFile: output file name
*/

void popData (char *line, 
				  int & lastUsed, 
				  int & inNamesPtr, 
				  int & outNamesPtr,
              int & latchUsed, 
				  int & latchInPtr, 
				  int & latchOutPtr);
/* 
	Function: Populate the data by calling the correct functions
	Variables:
	line: a line read from .blif file
	lastUsed: counter for the LUTs
	inNamesPtr: Pointer for the input names
	outNamesPtr: Pointer for the output name
	latchUsed: counter for the latches
	latchInPtr: Pointer for the input names of the latches
	latchOutPtr: Pointer for the output names of the latches
*/

int isLut (char *line);
/* 
	Function: Check if a LUT is described in the .blif file
	Variables:
	line: a line read from .blif file
*/

int isLatch (char *line);
/* 
	Function: Check if a latch is described in the .blif file
	Variables:
	line: a line read from .blif file
*/

void readBlif(FILE *infile, 
				  int functCall);
/* 
	Function: Read blif file and get a line from the file
	Variables:
	functCall: 0 = counting LUTs and Latches for memory allocation
	           1 = record data to the allocated memory arrays 
	
*/

void usage(char *arg);
/* 
	Function: Prompted message explained on how to use the program
	Variables: 
	arg: the name of the executable
*/

int parseArgs (int argc, 
					char *const *argv, 
               int &times, 
					double &error, 
					char *&inFile, 
					char *&outFile, 
					char *&netFile, 
					int &outType);
/* 
	Function: Load the user input arguments 
	Variables:
	argc: Number of arguments
	argv: arguments
	times: iteration on D-Flip-Flop Transition Density Calculation
	error: Error allowed for D-Flip-Flop Transition Density 
          Calculation
	inFile: input .blif file name
	outFile: log file name (*.txt)
	netFile: activity output file name (*.act)
	outType: 0 = output transition density only
	         1 = output probability only
				other (default) = output transition density 
				and probability 
				Note: VPR needs both transition density 
				      and probability values
*/

int find_FB_loop (int LUT_index,
						int* input_index);
/* 
	Function: Find all feedbackloops in the circuit

*/

void getLatchInputName (char *signal_name,
								int latch_index);
/* 
	Function: Get the input name of a latch
	Variables: 
	signal_name: signal name
	latch_index: Index for particular latch
*/					
void getLatchOutputName (char *signal_name,
								 int latch_index);
/*
  	Function: Get the output name of a latch
	Variables: 
	signal_name: signal name
	latch_index: Index for particular latch
 */

void getLUT_InputName (char *signal_name,
							  int LUT_index,
							  int input_index);
/*
  	Function: Get the input name of a LUT
	Variables: 
	signal_name: signal name
	LUT_index:   Index for the LUT
	input_index: Index for the LUT input
 */

void getLUT_OutputName (char *signal_name,
								int LUT_index);
/*
  	Function: Get the input name of a LUT
	Variables: 
	signal_name: signal name
	LUT_index:   Index for the LUT
 */


int filter (double *probability,
				 double *density);
/*
  Function: Filter Out the short pulses when the transition density 
            is higher than half of the clock (when the density is 
				higher than 1)
  Variables:
  probability = probability of a particular signal
  density = transition density of a particular signal
*/				 

/* Debugging functions */
void printLutOrder();
void printLutList();
void printArray(int inputArray[],
					 int arraySize);

void printDouble(double doubleArray[],
					  int arraySize);
void print_table ();

/***************************************************************************/
/* 
	Function: Build a dynamic table based on the LUT size 
*/
void build_table ()
{
  int function_number;
  int i;
  
  function_number = (int) pow(2,lut_size);

  table = new char* [function_number];

  for (i=0; i<function_number; i++)
	 {
		table[i] = new char[lut_size+1];

		int2str(i,
				  lut_size+1,
				  table[i]);
	 }
}
/***************************************************************************/
/* 
	Function: Delete the dynamic table based on the LUT size 
*/
void delete_table ()
{
  int function_number;
  int i;

  function_number = (int) pow(2,lut_size);
  for (i=0; i<function_number; i++)
	 {
		delete table[i];
	 } 
  delete table;

}
/***************************************************************************/
/* Print out the generated table for debugging */
void print_table ()
{
  int i;
  int function_number;

  function_number = (int) pow(2,lut_size);
  for (i=0; i<function_number; i++)
	 {
		printf ("Table:%s\n", table[i]);
	 } 
}

/***************************************************************************/
/* Allocate Memory for the LUT record */
void built_LUT_Data (int numLut)
{
    int function_number;
	 int i;
	 
	 function_number = (int) pow(2,lut_size);
	 
	 for (i=0; i<numLut; i++)
		{
		
		  luts[i].inChange = new double[lut_size];
		  luts[i].inDensity = new double[lut_size];
		  luts[i].inNameStart = new int[lut_size+1];
		  luts[i].outputHigh = new int[function_number];

		  /* Assume only one output per LUT */
		  luts[i].outProb = new double[2];
		  luts[i].outNameStart = new int[2];
		}
	 

}
/***************************************************************************/
/* Deallocate Memory for the Lut Record */
void delete_LUT_Data (int numLut)
{
	 int i;

 	 for (i=0; i<numLut; i++)
		{
		  delete [] luts[i].inChange;
		  delete [] luts[i].inDensity;
		  delete [] luts[i].inNameStart;
		  delete [] luts[i].outputHigh;
		  delete [] luts[i].outProb;
		  delete [] luts[i].outNameStart;
		} 
}

/***************************************************************************/
void int2str (int input_num,
				  int string_len,
				  char* outputstr)
{
  int temp_int;
  int remainder;
  char reverse_binary_str[20];
  char binary_str[20];
  int i,j, k, h;

  temp_int = input_num;  
  
  //printf ("temp_int: %d\n", temp_int);

  i =0;
  do {
	 remainder = temp_int % 2;
	 temp_int  = temp_int/2;

	 if (remainder == 1)
		reverse_binary_str[i++] = '1';
	 else
		reverse_binary_str[i++] = '0';

  } while (temp_int > 0);
  reverse_binary_str[i] ='\0';

  k = string_len-i-1;

  for (h=0; h<k; h++)
	 {
		binary_str[h] = '0'; 
	 }


  for (j=0; j<i; j++)
	 {
		binary_str[k+j] = reverse_binary_str[i-j-1];
	 }
  binary_str[k+j] = '\0';
  strcpy(outputstr, binary_str);
}

/***************************************************************************/
/* Allocate Memory for Luts and Latches */ 
void allocMem (int numLut, int numLatch)
{
    int i, k, l;    // Counter variables
	 int function_number;   

    numLuts = numLut;             // Record number of luts
    numLatches = numLatch;        // Record number of latches

    /*  allocate memory */
    linked = new int [numLut];
    luts = new LUTDATA [numLut];

	 printf("numLut:%d\n",numLut);
	 built_LUT_Data(numLut);

	 function_number = (int) pow(2,lut_size);

    printf ("Number of Luts: %d   Number of latches %d\n\n",numLut,numLatch);

    /*  Initialize various data structure values to invalid */ 
        
    for (i = 0 ; i < numLut; i++)
    {
        linked [i] = -1;          // Set to -1 to indicate not linked
        luts[i].numInputs = 0;    // Set the number of inputs to 0 first.

        for (l = 0; l <=lut_size; l++)
        {
            luts[i].inNameStart[l] = -1;  
        }
        for (k = 0; k<function_number; k++)
        {
            luts[i].outputHigh[k] = -1;
        }
        for (k = 0; k<2;k++)
        {
            luts[i].outProb[k] = -1;
        }
    }

    order = new int [numLut];
    inputNames[0] = '\0';
    outputNames[0] = '\0';
    
    latches = new LATCHDATA [numLatch]; 
    for (i = 0; i < numLatch; i++)
    {
        latches[i].inProb = PI_Prob;
        latches[i].outProb = PI_Prob;
    }
	

}

/***************************************************************************/
/*  Delete all dynamically allocated memory */
void freeMem ()
{
    delete [] linked;
	 delete_LUT_Data;
    delete [] luts;
    delete [] order;
    delete [] latches;
    
    netList.clear();

}

/***************************************************************************/
/*  Add a signal name and its probability value to the netList hash table  */
void addNet (char *buffer, 
				 double prob)
{
    string signal;
    STRING2F::iterator it; 
    
    // Record the net into the netlist hash table
    signal = (string)buffer;
    
    if (prob != -1)
    {
        if ((it = netList.find(signal)) != netList.end()) 
        {
            netList.erase (it);
        } 
    }

    netList.insert(STRING2F::value_type(signal,prob)); 

}


/***************************************************************************/
/*  Get the probability value based on the signal name, 
    by looking up the netList hash table */
double getProb (char *signal)
{
    STRING2F::iterator it;
    string mySignal;
    
    mySignal = (string)signal;
    if ((it = netList.find(mySignal)) != netList.end())
        return (double)(*it).second;
    else
    {
        printf ("ERROR signal name not in net list!\n");
        return -1;
    } 
}


/***************************************************************************/
/*  Add a input name to a lut  */
void addInName (char * buffer, 
		int startpos, 
                int endpos, 
		int inNamePtrPos,
                int lutIndex)
{
    char input [80];
    int inputIndex;
    string signal;
    
    inputIndex = luts[lutIndex].numInputs;
    
    //  Record the input Names into the buffer.
    strncpy (input, buffer+startpos, endpos-startpos-1);
    input[endpos-startpos-1] = '\0';
    strcpy (inputNames+inNamePtrPos,input);
    inputNames[inNamePtrPos+(endpos-startpos-1)] = '\0';

    
    // Record the location of the input names.
    luts[lutIndex].inNameStart[inputIndex] = inNamePtrPos;
    luts[lutIndex].inNameStart[inputIndex+1] = inNamePtrPos + (endpos - startpos - 1);
    
    luts[lutIndex].numInputs++;
    
    addNet (input, -1);

}

/***************************************************************************/
/*  Add a output name to a lut  */
void addOutName (char * buffer, 
					  int startpos, 
                 int endpos, 
					  int outNamePtrPos,
                 int lutIndex)
{
    char output [80];
    string signal;
    
    //  Record the output Names into the buffer.
    strncpy (output, buffer+startpos, endpos-startpos-1);
    output[endpos-startpos-1] = '\0';
    strcpy (outputNames+outNamePtrPos,output);
    outputNames[outNamePtrPos+(endpos-startpos-1)] = '\0';

    
    // Record the location of the input names.
    luts[lutIndex].outNameStart[0] = outNamePtrPos;
    luts[lutIndex].outNameStart[1] = outNamePtrPos + (endpos - startpos - 1);

    addNet (output, -1);

}

/***************************************************************************/
/*  Add a input or output name to a latch  */
void addLatch (char *buffer, 
					int startpos, 
					int endpos, 
					int ptrPos, 
               int index, 
					int in_out)
{
    
    char value [80];
    
    //  Record the Name into the buffer.
    strncpy (value, buffer+startpos, endpos-startpos-1);
    value[endpos-startpos-1] = '\0';

    if (in_out ==1) //in
    { 
        strcpy (latch_in+ptrPos,value);
        latch_in[ptrPos+(endpos-startpos-1)] = '\0';
 
        // Record the location of the input names.
        latches[index].inNameStart[0] = ptrPos;
        latches[index].inNameStart[1] = ptrPos + (endpos - startpos - 1);
    }
    
    
    if (in_out ==2) //out
    { 
        strcpy (latch_out+ptrPos,value);
        latch_out[ptrPos+(endpos-startpos-1)] = '\0';
 
        
        // Record the location of the output names.
        latches[index].outNameStart[0] = ptrPos;
        latches[index].outNameStart[1] = ptrPos + (endpos - startpos - 1);
    }
}

/***************************************************************************/
/*  Add input name to the netList hash table */
void addInput (char *buffer, 
	       int startpos, 
	       int endpos)
{
    
    char value [80];
    string signal;
    double initialDensity;
    
    //  Record the Name into the buffer.
    strncpy (value, buffer+startpos, endpos-startpos-1);
    value[endpos-startpos-1] = '\0';

    addNet (value, PI_Prob);


	 initialDensity = PI_Density;

}

/***************************************************************************/
/*  Compute the "truth table" of the lut  
    the "truth table is stored in luts[i].outputHigh[j]
    a 1 means the binary input of "j" is high
    and a 0 means it is low.
*/ 
void addHighEntry (char * buffer, 
						 int lutIndex)
{    
    int i, j, k,l;
    char target;
 
  	 int *valid;
	 int function_number;  

	 function_number = (int) pow(2,lut_size);
	 valid = new int[function_number];

    for (i = 0; i < function_number; i++)
    {
        valid[i] = 0;
    }
    
    l = strlen (buffer) -3;

    k = (int)pow (2, (strlen(buffer) -4)) ;
    k = k*2 -1;

    
    for ( i =0; i < l; i++)
    {
        if (buffer[i] == '1')
        {
            target = '1'; 
        }
        else if (buffer[i] == '0')
        {
            target = '0';
        }
        else if (buffer[i] == '-')
        {
            target = '2';
        }
        
        for (j = 0; j<=k; j++)
        {
            if ( (table[j][lut_size-l+i] == target) && (valid[j] != -1) )
            {
                valid[j] = 1;
            }
            else if ( (target =='2') && (valid[j] != -1) )
            {
                valid[j] = 1;
            }
            else if (table[j][lut_size-l+i] != target)
            {
                valid[j] = -1;
            }
        }
    }
    
    for (j = 0; j<=k; j++)
    {
        if (valid[j] == 1)
        {
            luts[lutIndex].outputHigh[j] = 1;
        }
        else if (luts[lutIndex].outputHigh[j] != 1) 
        {
            luts[lutIndex].outputHigh[j] = 0;
        }
    }
	 delete valid;
}

/***************************************************************************/
/*  Compute the order of the Luts to compute */
void orderLuts()
{
    /* Array pointer variable that points to the 
       starting and ending position of various signals 
		 (does not include inputs or outputs)
    */
    int * signalStart ;            
 
    char signalNames[MAXSIZE];             /*  Signal name storage buffer */
    signalStart = new int [numLuts * lut_size];
    
    int signalNamePtrPos = 0;              /*  Index pointer to tell where 
															  to write data to */
    
    int numLinked = 0;                     /*  Number of luts that have 
															  been ordered */
    int okLink = 0;                        /*  Boolean variable that stops loop */
    int i, j, k;                           /*  Counter variables */
    int inputsValid;                       /*  Boolean variable that determines 
															  if the inputs of the Lut are "valid"
															  for ordering */
    char targetSignal[20 * lut_size];                 /*  signal Name that should be check to 
															  see if its output has been ordered */ 
    char signalName[20 * lut_size];                   /*  signalName used to compare with 
															  targetSignal */
    char output[20*lut_size];                       /*  Records the output name of the lut  */ 
    
    int numSignal = 0;                     /*  Number of signals for a particular Lut */
    double prob;                           /*  Record the probability  */
    
    while (numLinked != numLuts)
    {
        printf ("     %d of %d luts have been linked.\n",numLinked,numLuts);
        for (i = 0; i < numLuts; i++)   // Loop through the Luts
        {
            okLink = 0;
            inputsValid = 1;
            if (linked[i] != 1)   // The lut has not been linked
            {
                j=0;
                
                while ( (inputsValid) && (j < luts[i].numInputs) )
                {
						okLink = 0;

						/* Get the name of the input */
						strncpy (targetSignal, inputNames+luts[i].inNameStart[j],
									luts[i].inNameStart[j+1] - luts[i].inNameStart[j] );
						targetSignal [ luts[i].inNameStart[j+1] - luts[i].inNameStart[j] ] = '\0';
						
						/* Get the probability of the input */
						prob = getProb (targetSignal);
						
						if (prob != PI_Prob)  // Not a input or latch output. 
                    {
							 /* Check and see if the primary inputs or the output of
								 other luts are the inputs of this lut */
							 for (k = 0; k <numSignal; k++)
                        {
								  strncpy (signalName, signalNames+signalStart[k],
											  signalStart[k+1] - signalStart[k] );
								  signalName[signalStart[k+1] - signalStart[k]]='\0';
								  
								  /* if The input matches the output of LUTs */
								  if (strcmp (signalName, targetSignal) == 0)
                            {
										okLink = 1;
										k = numSignal;  // Exit loop for efficiency.
                            }
                        }// for
                    } // if prob 
                    else //Input
                    {
                        okLink =1; 
                    }
                    
                    if (okLink != 1) //Input is invalid
                    {
                        inputsValid = 0;
                    }
                    else
                    {
                        j++;
                    } 
                    
                } //While
        
                if (luts[i].numInputs == 0)   // Special no input LUT case.
                    okLink = 1;
                
					 /* Check if the lut has been successfully linked */
                if (okLink == 1)
                {
                    linked[i] = 1;
                    order[numLinked] = i;
                    numLinked++;
                    
						  /* Record the output of this lut to the list of possible input signals
							  for another LUTs */
                    strncpy (output, outputNames+luts[i].outNameStart[0],
                        luts[i].outNameStart[1] - luts[i].outNameStart[0] );
                    output [ luts[i].outNameStart[1] - luts[i].outNameStart[0] ] = '\0';
                    
                    strcpy (signalNames+signalNamePtrPos, output);   
                    signalNames[signalNamePtrPos+ 
                        (luts[i].outNameStart[1] - luts[i].outNameStart[0] )] = '\0';
                    signalStart[numSignal] = signalNamePtrPos;
                    signalStart[numSignal+1] = signalNamePtrPos+ 
                        (luts[i].outNameStart[1] - luts[i].outNameStart[0] ); 

                    numSignal++;
                    signalNamePtrPos = 
							 signalNamePtrPos + (luts[i].outNameStart[1] - luts[i].outNameStart[0] ) ;
                }
            }  // If lut has not been linked
        } // For loop through the luts
    }  //While unlinked.

    /* Clean up the allocated Memory */
    delete [] signalStart;
    printf ("\n");
}

/*  Calculate the probability that the output is high for each LUT */
int calcProb (double err)
{
    
    int i,j,k,l,m,o;       /*  Counter values */
    int equal;             /*  boolean value */
    double probValue;      /*  Single probability value for input */
    double totalProb;      /*  Total probability value for output */
    int targetLut;         /*  Lut to do  */
    int inputsValid;       /*  boolean value */
    char input[80];        /*  Input name */
    char output [80];      /*  Output name */
    char latchName[80];    /*  Latch input name */
    double error;          /*  Error value to tell when to stop iterations */
    int cont;              /*  boolean value for termination of loop */
    int counter = 0; 
    int * done;            /*  Flag variable that checks if latch values have been propogated */
    int prop;              /*  boolean value to determines if latch values should be propogated */     
    double *prob;        /*  Probability values of the inputs */
	 double max_difference; /*  Maximum Difference on probability calculation */
	 double average_difference; /* Average Difference on probability calculation */
    double total_difference;   /* Total Difference on probability calculation */
    

	 prob = new double[lut_size];
    done = new int [numLatches];
    
    
    for (i = 0; i < numLuts; i++)   // Loop through the number of Luts.
    {
        
        targetLut = order[i];    // Get lut to compute
        inputsValid = 1;         // Set boolean variable
        totalProb = 0;           // reset probability
        
        
        /*  Get the probability values of all the inputs and
            store them in prob[j]
        */
        for (j = 0; j<luts[targetLut].numInputs; j++)
        {
			 getLUT_InputName(input, targetLut, j);
			 prob[j] = getProb (input);
        }
        
        /*  Determine number of entries in truth table */
        l = 2 * ((int) pow(2,luts[targetLut].numInputs-1) );
        /*  Compute the probability output is high based on truth table 
        and input probability values */
        if (luts[targetLut].numInputs == 0)
            totalProb = luts[targetLut].outputHigh[0];
        else
        {
			 // Loop through the number of entries in the truth table
            for (k = 0; k<l; k++)    
            {
                if (luts[targetLut].outputHigh[k] == 1)
                {
                    probValue = 1;
                    j = 0;
                    
                    for (m = lut_size-luts[targetLut].numInputs; m<lut_size; m++)
                    {
                        if (table[k][m] == '1')
                        {
                            probValue = probValue * prob[j];
                        }
                        else
                        {
                            probValue = probValue * ( 1- prob[j]);
                        }
                        j++;
                    }  
                    totalProb = totalProb + probValue;
                }
            }  // for number of entries in truth table
        }   
        luts[targetLut].outProb[1] = totalProb;     /* Record the probablility */ 
        
		  getLUT_OutputName(output, targetLut);
	
        addNet (output, totalProb);

		  /* check if output name is actually a latch input
			  if true, record the probablity values into the latch
		  */ 
		  for (o=0; o<numLatches; o++)
				  {
					 getLatchInputName(latchName, o);
					 
					 equal =strcmp (latchName, output); 
					 if (equal ==0)
						{
						  
						  latches[o].inProb = totalProb;
						  latches[o].outProb = totalProb;
						  done[counter] = o;
						  counter++;
						}
				  }      
		  
	 }  // For number of Luts
	 
	 // No iteration is required for combination circuits
	 if (numLatches == 0)
		{
		  delete [] prob;
		  delete [] done;
		  return 1; // noError.
		}
	 
	 // For sequential circuits,
	 // Check if iteration should continue or not based on the error desired. 
	 max_difference = 0;
	 average_difference =0;
	 total_difference = 0;

    cont =0;
    for (i = 0; i< numLuts; i++)
    {
        error = (luts[i].outProb[1]-luts[i].outProb[0]);
        if (error < 0)
            error = -1 * error;
        if (error >err)
        {
            cont = 1;
        } 
		  
		  if (max_difference< error)
			 {
				max_difference = error;
			 }
		  total_difference += error;		  
    }
    
	 average_difference = total_difference/numLuts;
	 
	 printf ("For this iteration\n");
	 printf ("average_difference: %f\n", average_difference);
	 printf ("max_difference: %f\n", max_difference);

    if (cont == 1)
    {
        // Setup for next iteration.
        
        // Backup the previous output probability value for future comparisions
        for (i = 0; i< numLuts; i++)
        {
            luts[i].outProb[0] = luts[i].outProb[1];
        }
        
        /*  Propogate the latch values in -> out */
        for (k = 0; k < numLatches; k++)
        {
            prop = 0;
            for (j = 0; j<counter; j++)
            {
                if (k == done[j])
                {
                    prop = 1;
                }
            }
            if (prop == 0)
            {
                latches[k].outProb = latches[k].inProb;
            }
        }
        
        
        for (k = 0; k < numLatches; k++)
        {
			 getLatchOutputName(latchName, k);
			 
			 for (l=0; l<numLatches; l++)
            {
				  getLatchInputName(input, l);

				  if (strcmp (input, latchName) == 0)
                {
						latches[l].inProb = latches[k].outProb;
                }
            }
			 
			 addNet (latchName, latches[k].outProb);
        }
        delete [] prob;
        delete [] done;
        return 0; // Error.
    }
    else
    { 
		delete [] prob;
		delete [] done;
		return 1; //no error.
    }
    
}



 
/***************************************************************************/
/*  Calculate the transition probability, which is 2p(p-1), for the inputs
    and output of the lut  */
void calcChange()
{
    int i, j;
    char input[80];
    double prob;
    
    for (i =0; i < numLuts; i++)
    {
        for (j=0; j<luts[i].numInputs; j++)
        {
			 /* Get the name of the lut input */
			 strncpy (input, inputNames+luts[i].inNameStart[j],
						 luts[i].inNameStart[j+1] - luts[i].inNameStart[j] );
			 input[ luts[i].inNameStart[j+1] - luts[i].inNameStart[j] ] = '\0';
			 
			 /* Get the Probability of the input */
			 prob = getProb (input);
			 
			 /* Calculate the Transition Probability of the lut input */
			 luts[i].inChange[j] = 2*prob*(1-prob);
        }

		   /* Calculate the Transition Probability of the lut output */
        luts[i].outChange = 2*luts[i].outProb[1]*(1-luts[i].outProb[1]);
    }
}

/***************************************************************************/
/*  Calculate the transition density */
void calcDensity()
{
    int i, j, k, l, m, n, p, q, s;    /* counter variables */
    char input[80];                   /* input name */
    char output[80];                  /* output name */
    char inputSignal[80];             /* input signal name */
    int targetLut, targetLut2;        /* lut to compute */
	 double temp_LUT_PROB;             /* temporary storage for the
													  probability */
	 int FB_index;                     /*  Index for a feedback intput  */									  
	 
	 /* Initialization */
    for (i = 0 ; i < numLuts; i++)
    {
		for (j = 0 ; j < luts[i].numInputs; j++)
        {
			 /* Initialize the transition density 
			    to the user input value*/
			 luts[i].inDensity[j] = PI_Density;
        }

		/* 
			initialize for feedback loop to
			the transition probability
		 */
		if (find_FB_loop(i, &FB_index) == 1)
		{
		  luts[i].inDensity[FB_index] 
		    = luts[i].inChange[FB_index];
		}
    }
    

	 /* Transition Density Calculation */
    for (i = 0; i < numLuts; i++)
    {
        targetLut = order[i];
		  calcDensity_perLUT(targetLut);
		  getLUT_OutputName(output, targetLut);
		  temp_LUT_PROB = getProb(output);

		  if (filter (&temp_LUT_PROB, &luts[targetLut].outDensity))
		  {
			 luts[targetLut].outProb[1] = temp_LUT_PROB;
			 luts[targetLut].outChange = 2*temp_LUT_PROB*(1-temp_LUT_PROB);
			 addNet(output, temp_LUT_PROB);
		  }
		  		  

		  /* Loop through all the outputs of the LUTs and check if
			  the lut output serves as the input of other luts. 
			  if true, modify their input Transition Density.*/
		  
		  
		  for (n = i+1; n<numLuts; n++)  
			 {
				targetLut2 = order[n];
				j = 0;
				
				while ( j <luts[targetLut2].numInputs) 
				  {
					 
					 /* get the name of the inputSignal*/
					 
					 getLUT_InputName (inputSignal, targetLut2, j);
					 
					 if (strcmp (output, inputSignal) == 0)
						{
						  luts[targetLut2].inDensity[j] = luts[targetLut].outDensity; 
						  
						  addNet(inputSignal, luts[targetLut].outProb[1]);
						}
					 j++;
				  }  
			 }
		 
 
	 }
}

/***************************************************************************/
void calcDensity_perLUT (int targetLut)
{

  int l, m, n, k, p, q,s;           /* counter variables */
  double totalDensity;              /* storage variable to calculate 
													density */ 
  int *xorTable;
  int xorTable_size;
//int xorTable[8];                  /* xorTable for transition density */
  int offset;                       /* adjustment variable */
  
  double *prob;
  //double prob[4];                   /* probability values */
  double probValue;                 /* storage variable to calculate 
													  density */
  double density;                   /* density value */
  double totalProb;                 /* storage variable to calculate 
													probabiltiy */
  char input[80];                   /* input name */
  char output[80];                  /* output name */


  xorTable_size = (int) pow(2, lut_size-1);
  xorTable = new int[xorTable_size];
  prob = new double[lut_size];

  totalDensity = 0;


  if (luts[targetLut].numInputs == 0)
	 {
		/* LUT with no inputs */
		totalDensity = 0;
		
	 }
  else
	 {
		/* Calculate the differentiation of the logic function implemented */
		/* by building a XOR table for the LUT */
		
		/* calculate how many logical combinations based on
			the number of LUT inputs */
		l = 2* ((int) pow(2,luts[targetLut].numInputs-1) );
		
		for (k = 0; k<luts[targetLut].numInputs; k++)
		  {
			 s = 0;
			 
			 /* load the probabilities of the lut input */
			 for (p =0; p < luts[targetLut].numInputs; p++)
				{
				  if (p != k)
					 {
						/* get input name */
						getLUT_InputName(input,targetLut,p);
						
						/* get probability of the input */
						prob[s] = getProb(input);
						s++;
					 }
				}
			 
			 /* For each input, two Boolean functions are generated 
				 by setting the input to 0 and 1.*/
			 n=0;
			 
			 /* For each Boolean function, a truth table is set up*/
			 /* An offset value is generated to compare the values of
				 two Boolean functions */ 
			 offset = (int) pow (2, luts[targetLut].numInputs-1-k);
			 
			 /* Build the XOR function based on the two Boolean 
				 functions */
			 for (m=0; m<l; m++)
				{
				  if (table[m][lut_size-luts[targetLut].numInputs+k] == '0')
					 {
						xorTable[n] = 
						  luts[targetLut].outputHigh[m] ^ luts[targetLut].outputHigh[m+offset];
						n++;
					 }
				}
			 
			 /* Loop through the xor truth table and calculate the
				 static probability contributed by each LUT input */
			 
			 
			 totalProb = 0;
			 for (p =0; p<n; p++) 
				{
				  if (xorTable[p] == 1) //If table is 1
					 {   
						probValue = 1;
						q=0;
						// loop through the bits of each entry in the Xor table
						for (s=lut_size-luts[targetLut].numInputs+1; s<lut_size; s++) 
						  {
							 if (table[p][s] == '1')
								{
								  probValue = probValue * prob[q];
								}
							 else
								{
								  probValue = probValue * (1-prob[q]);
								}
							 q++;
						  } 
						totalProb = totalProb + probValue;
						
					 }  // If table is 1
				}  // Loop through the xor truth table
			 density = totalProb * luts[targetLut].inDensity[k];
			 totalDensity = totalDensity + density;
			 
		  } // for k<numInputs
		
	 } // else
  
  
  
  
  /* Record the Transition Density for the LUT output */
  luts[targetLut].outDensity = totalDensity; 

  delete [] xorTable;
  delete [] prob;
  
}

/***************************************************************************/
/* Get the transition density value based on the signal name */
double getDensity (char *signal)
{

  int i, j;               /* counter variable */
  int found ;             /* boolean function indicate the signal 
									  is found */ 
  double density = -1;    /* density value */
  char inputSignal[80];   /* character array for storing the LUT 
									  input name */
  char latchName [80];    /* character array for storing the latch
									  output name */
  char output [80];       /* character array for storign the LUT output
									  name */

  /* Initialization */
  i = 0;
  found = 0;

  /* Loop through all the luts */
  while ( (i<numLuts) && (found !=1) ) 
	 {
      j= 0;
		while( j <luts[i].numInputs && (found != 1) ) 
		  {
			 /* get the lut input name */
			 strncpy (inputSignal, inputNames+luts[i].inNameStart[j],
						 luts[i].inNameStart[j+1] - 
						 luts[i].inNameStart[j] );
			 inputSignal [ luts[i].inNameStart[j+1] - 
							 luts[i].inNameStart[j] ] = '\0';
			 
			 /* compare the input name with the signal name*/
			 if (strcmp (signal, inputSignal) == 0)
				{
				  density = luts[i].inDensity[j]; 
				  found = 1;
				}
			 j++;                
		  }
		
		/* get the lut output name */
		strncpy (output, outputNames+luts[i].outNameStart[0],
					luts[i].outNameStart[1] - luts[i].outNameStart[0] );
		output [ luts[i].outNameStart[1] - luts[i].outNameStart[0] ] = '\0';
		
		/* compare the lut output name with the signal name */
		if (strcmp (signal, output) == 0)
		  {
			 density = luts[i].outDensity; 
			 found = 1;
		  }
		
		i++;
	 }
  
  /* Loop through all the latches */
  i = 0;
  while ( ( i<numLatches) && (found !=1) )
	 {
		/* get latch name */
      strncpy (latchName, latch_out+latches[i].outNameStart[0],
               latches[i].outNameStart[1] - latches[i].outNameStart[0] );
		latchName [ latches[i].outNameStart[1] - latches[i].outNameStart[0] ] = '\0';
		
		/* compare latch name with the signal name */
		if (strcmp (signal, latchName) == 0)
		  {
			 density = latches[i].outDensity;
			 found = 1;
		  }
		i++;
	 } 

  if (found == 1)
	 return density;
  else
	 /* if not found, assume that is a global clock signal */
	 return 2; // clock.
}

/***************************************************************************/
/* dump values to file */
void dumpNetList (char *outFile, 
						int outType)
{
    STRING2F::iterator it;
    int length;
    char signal[80];
    double prob, density;
	char buffer[1000];

	FILE *stream;
    
    if ( !(stream = fopen (outFile, "w" )))
    {
        fprintf (stderr, "unable to open %s for writing", outFile);
        exit (2);
    }

	printf ("\nNet list has been written to the file %s\n", outFile);

	for (it = netList.begin (); it != netList.end(); it++)
    {
        length = (*it).first.length() ;
        (*it).first.copy (signal, length);
        signal [length] = '\0'; 
        prob = (*it).second;
        sprintf (buffer,"%s ",signal);
		fwrite (buffer, sizeof (char), strlen (buffer), stream);
		density = getDensity (signal);

		if (outType == 1)
		{
			sprintf (buffer,"%f\n", prob);
			fwrite (buffer, sizeof (char), strlen (buffer), stream);
		}
		else if (outType == 0)
		{
			sprintf (buffer,"%f\n", density);
			fwrite (buffer, sizeof (char), strlen (buffer), stream);
		}
		else
		{
			sprintf (buffer,"%f %f\n", prob, density);
			fwrite (buffer, sizeof (char), strlen (buffer), stream);
		}

    }
    
    fclose(stream);
    
    
}



/***************************************************************************/
/* Write log data to *.txt file */
void dump2File (char *outFile)
{
    int i, j, k,l;
    char buffer[1000];
    char input [80];
    char output [80];
    char latchNameIn[80];
    char latchNameOut[80];
	 int function_number;
		
	 function_number = (int) pow(2,lut_size);

    FILE *stream;
    
    if ( !(stream = fopen (outFile, "w" )))
    {
        fprintf (stderr, "unable to open %s for writing", outFile);
        exit (2);
    }
    
    printf ("Probability values have been written to the file %s\n", outFile);
    sprintf (buffer, "Number of Luts: %d\n", numLuts);
    fwrite (buffer, sizeof (char), strlen (buffer), stream); 
    
    
    for (i = 0 ; i < numLuts; i++)
    {
        // Print out the input names and their probibilty values.
        sprintf (buffer,"\n\nLut %d\n", i);
        fwrite (buffer, sizeof (char), strlen (buffer), stream); 
        k = 0;
        l = 0;
        
        for (j = 0 ; j < luts[i].numInputs; j++)
        {
            strncpy (input, inputNames+luts[i].inNameStart[j],
                luts[i].inNameStart[j+1]-luts[i].inNameStart[j] );
            input[ luts[i].inNameStart[j+1]-luts[i].inNameStart[j] ] = '\0';
            
            sprintf (buffer,"Input: %s\n", input);
            fwrite (buffer, sizeof (char), strlen (buffer), stream); 
            sprintf (buffer,"   Static Probability:     %f\n", getProb(input));
            fwrite (buffer, sizeof (char), strlen (buffer), stream); 
            sprintf (buffer,"   Transition Probability: %f\n" ,luts[i].inChange[j]);
            fwrite (buffer, sizeof (char), strlen (buffer), stream); 
            sprintf (buffer,"   Transition Density:     %f\n",luts[i].inDensity[j]);
            fwrite (buffer, sizeof (char), strlen (buffer), stream); 

        }
        
        
        // Print out the output names and their probability values.
        sprintf (buffer,"\n");
        fwrite (buffer, sizeof (char), strlen (buffer), stream); 
        strncpy (output, outputNames+luts[i].outNameStart[0], 
            luts[i].outNameStart[1]-luts[i].outNameStart[0] ); 
        output[ luts[i].outNameStart[1]-luts[i].outNameStart[0] ] = '\0'; 
        
        sprintf (buffer, "Output: %s\n", output);
        fwrite (buffer, sizeof (char), strlen (buffer), stream); 
        sprintf (buffer, "   Static Probability:     %f\n", luts[i].outProb[1]);
        fwrite (buffer, sizeof (char), strlen (buffer), stream); 
        sprintf (buffer, "   Transition Probability: %f\n", luts[i].outChange);
        fwrite (buffer, sizeof (char), strlen (buffer), stream); 
        sprintf (buffer, "   Transition Density:     %f\n", luts[i].outDensity);   
        fwrite (buffer, sizeof (char), strlen (buffer), stream); 

        
        // Print out the truth table of the LUTS.
        sprintf (buffer,"\n");
        fwrite (buffer, sizeof (char), strlen (buffer), stream); 
        sprintf (buffer,"Output is high for values: ");
        fwrite (buffer, sizeof (char), strlen (buffer), stream); 
        while ((!l) && k <function_number)
        {
            if (luts[i].outputHigh[k] == 1)
            {
                sprintf (buffer,"%d ",k);
                fwrite (buffer, sizeof (char), strlen (buffer), stream); 
            }
            else if (luts[i].outputHigh[k] == -1) 
            {
                l = 1;
            }
            k++;
        }
        sprintf (buffer,"\n");
        fwrite (buffer, sizeof (char), strlen (buffer), stream); 
        
    }
    
    // Print out the order of the luts to be calculated.
    sprintf (buffer,"\n");
    fwrite (buffer, sizeof (char), strlen (buffer), stream); 
    for (i = 0 ; i<numLuts; i++)
    {
        sprintf (buffer,"Order of luts to be done %d is lut %d\n",i,order[i]);
        fwrite (buffer, sizeof (char), strlen (buffer), stream); 
    }   
    
    // Printf the latch names.
    if (numLatches >0)
    {
        sprintf (buffer,"\nNow printing out the latch names and probability values.\n");
        fwrite (buffer, sizeof (char), strlen (buffer), stream); 
        for (i = 0; i < numLatches; i++)
        {
            strncpy (latchNameIn, latch_in+latches[i].inNameStart[0],
                latches[i].inNameStart[1] - latches[i].inNameStart[0] );
            latchNameIn [latches[i].inNameStart[1] - latches[i].inNameStart[0]] = '\0';
            
            strncpy (latchNameOut, latch_out+latches[i].outNameStart[0],
                latches[i].outNameStart[1] - latches[i].outNameStart[0] );
            latchNameOut [latches[i].outNameStart[1] - latches[i].outNameStart[0]] = '\0';
            
            sprintf (buffer,"latchNameIn: %s latchNameOut: %s inProb: %f outProb %f\n",
                latchNameIn, latchNameOut, latches[i].inProb, latches[i].outProb);
            fwrite (buffer, sizeof (char), strlen (buffer), stream); 
        }
    } 
    fclose(stream);
}

/***************************************************************************/
/* Populate the data by calling the correct functions */
void popData (char *line, 
				  int & lastUsed, 
				  int & inNamesPtr, 
				  int & outNamesPtr,
				  int & latchUsed, 
				  int & latchInPtr, 
				  int & latchOutPtr)
{   
    char *pdest;
    char *pdest2 = line;
    
	 /* String Pattern can be found in the .blif file */
    char str[] = ".names"; /*logic gate */
    char str2[] = " ";     /*separator between input and output names*/ 
    char str3[] = ".";     /*first character of lines*/
    char str4[] = ".latch";/*represents latches */
    char str5[] = ".inputs";/*represents primary input*/
    
    int  ptr = 0;
    int  ptr2 = 0;
    int  eoln = 0;
    
	 
    /*extract Logic gate descriptions*/
    pdest = strstr (line, str);   
    if ( pdest != NULL )
    {
        lastUsed++;
        ptr = 7;
        while (pdest2 != NULL)
        {
            pdest2 = strstr (line+ptr, str2);
            if (pdest2 == NULL)
                ptr2 = strlen (line);
            else
                ptr2 = pdest2 - line +1;
            
            if (pdest2 != NULL)
            {
				  /* record the inputs of the LUT */
				  addInName (line, ptr, ptr2, inNamesPtr, lastUsed);
				  inNamesPtr = inNamesPtr + (ptr2 - ptr -1);
            }
            else if (pdest2 == NULL)  //output
            {
				  /* record the output of the LUT */
				  addOutName(line, ptr, ptr2, outNamesPtr, lastUsed);
				  outNamesPtr = outNamesPtr + (ptr2 - ptr -1);

				  /* Special case where no input Lut.*/
				  if (luts[lastUsed].numInputs == 0) 
					 luts[lastUsed].outputHigh[0] = 1;
            }
            ptr = ptr2;
            
        } //while 
    } // if
    
    /* Special Case where the LUT has fixed output value */
    pdest = strstr (line, str3);    //  the line has .
    if (pdest == NULL)
    {
        if (line[strlen(line)-2] == '1')  //  Double check.
        {
            addHighEntry (line, lastUsed);
        }  
    } // if
    else // line has .
    {

		/* extract latch descriptions */
        pdest = strstr (line,str4);
        if (pdest != NULL)  // Got a .latch
        {
            latchUsed++;
            ptr = 6;
            
				/* skip the blank space */
            while (line[ptr] == ' ')
                ptr++;            
            
				/* extract the input of the latch */
            pdest2 = strstr (line+ptr,str2);
            ptr2 = pdest2 - line +1;
            addLatch (line,ptr,ptr2,latchInPtr,latchUsed,1);
            latchInPtr = latchInPtr + (ptr2 - ptr -1);
            ptr = ptr2;
            
				/* extract the output of the latch */
            pdest2 = strstr (line+ptr,str2);
            ptr2 = pdest2 - line +1;

				/* record the latch information to the array */
            addLatch (line,ptr,ptr2,latchOutPtr,latchUsed,2);
            addInput (line,ptr,ptr2);

				/* move the latch output pointer */
            latchOutPtr = latchOutPtr + (ptr2 - ptr -1);
            ptr = ptr2;
            
            
        }
        else
        {
			 /* extract primary input description */
			 pdest = strstr (line, str5);
			 if (pdest != NULL) // Got .inputs
            {
                
				  ptr = 7;
				  
				  /* skip the blank spaces */
				  while (line[ptr] == ' ')
					 ptr++;
				  pdest2 = line +ptr;

				  while (pdest2 != NULL)
                {
                    
                    pdest2 = strstr (line+ptr,str2);
                    if (pdest2 == NULL)
                        ptr2 = strlen (line);
                    else
                        ptr2 = pdest2 - line +1;
						  
						  /* record the primary input name */
                    addInput (line,ptr,ptr2);
                    ptr = ptr2;
                    
                }
            } 
        }
        
    }
}

/***************************************************************************/
/* Check if a LUT is described in the .blif file */
int isLut (char *line)
{
    
    char *pdest;
    
    char str[] = ".names";
    pdest = strstr (line, str);
    if ( pdest != NULL )
    {
        return 1;
    }
    return 0; 
}

/***************************************************************************/
/* Check if a latch is described in the .blif file */
int isLatch (char *line)
{
    
    char *pdest;
    
    char str[] = ".latch";
    pdest = strstr (line, str);
    if ( pdest != NULL )
    {
        return 1;
    }
    return 0; 
}

/***************************************************************************/
/* Read blif file and get a line */
void readBlif(FILE *infile, 
				  int functCall)
{
    int eoln = 0;
    int eof = 0;
    int i = 0;
    char character;
    char line[8192];  //8K line!
    int numLuts = 0;
    int numLatches = 0;
    
    int lastUsed = -1;
    int latchUsed = -1;
    int inNamePtr = 0;
    int outNamePtr = 0;
    int latchInPtr = 0;
    int latchOutPtr = 0;
    

	 /* Go through the blif file and count the number of LUTs and latches*/ 
    while (!eof)
    {
        eoln = 0;
        while (!eoln)
        {
            if (fread (&character, sizeof (char), (size_t) 1, infile) != 1)
            {
                eof = 1;
            }
            
            if ( (character == '\n') && (line[i-1] == '\\') ) 
            {
                i--;
            }
            else 
                
                if ( (character == '\n') && (line[i-1] != '\\') )
                {
                    eoln = 1 ;
                    line[i] = '\n';
                    line[i+1] = '\0';
                    i = 0;
                }
                else
                {
                    line[i] = character;
                    i++;
                }
        } // !eoln
        
  
        if (functCall == 0)
        {
			 /* Counting Phase */
			 /* record the number of LUTs and latches*/
            if ( isLut (line) == 1 )
                numLuts++;
            
            if ( isLatch (line) == 1 )
                numLatches++;
        }
        else if (functCall == 1)
        {
			 /* Recording Phase */
			 /* record the logical functions implemented in the circuit */
            popData (line, lastUsed, inNamePtr, outNamePtr, 
                latchUsed, latchInPtr, latchOutPtr); 
        }
    } //!eof
    
    if (functCall == 0)
    {
		/* allocate Memory for LUTs and Latches */
		allocMem (numLuts,numLatches); 
        
    }
}

/***************************************************************************/
/* Prompted message explained on how to use the program */
void usage(char *arg)
{
  printf ("Usage: %s \n", arg);
  printf ("-e [error value 0<e<1 (default: 0.05)] \n");
  printf ("-i [max number of iterations (default: 10)] \n");
  printf ("-f [input filename] \n");
  printf ("-o [output filename (default: activity.txt)] \n");
  printf ("-n [net list filename] (default: activity.act) \n");
  printf ("-d (output transition density (default)) | -p (output static probability) (default: -d -p)\n");
  printf ("-s [static probability of your primary inputs (default: 0.5)\n");
  printf ("-t [transition density of your primary inputs (default: 0.5)\n");
  printf ("-l [look-up-table size] (default: 4)\n");
  printf ("-r [rise and fall time percentage relative to the clock period (default: 0.1 = 10 percent)\n");
}

/***************************************************************************/
/* Load the user input arguments */
int parseArgs (int argc, 
	       char *const *argv, 
               int &times, 
	       double &error, 
	       char *&inFile, 
	       char *&outFile, 
	       char *&netFile, 
	       int &outType)
{
    int i;
    
    int command;
    int dash = '-';
    char option;
    
    i =0;
    while ( (i<argc) || (command == 1) )
    {
        if (command ==1)
        {
            switch (option)
            {
            case 'e':
                //printf ("option e, with value %s\n",argv[i]);
                error = atof (argv[i]);
                break;
                
            case 'i':
                //printf ("option i, with value %s\n",argv[i]);
                times = atoi(argv[i]);
                break;
                
            case 'f':
                //printf ("option f, with value %s\n",argv[i]);
                inFile = argv[i];
                break;
                
            case 'o':
                //printf ("option o, with value %s\n",argv[i]);
                outFile = argv[i];
                break;

				case 'n':
				  netFile = argv[i];
				  break;

				case 'd':
				  if (outType == 2)  //user just entered -d
					 {
						outType = 0;
					 }
				  else if (outType == 1)  //user has entered -p
					 {
						outType = 2;
					 }
				  break;
				  
				case 'p':
				  if (outType == 2)  //user just entered -p
					 {
						outType = 1;
					 }
				  else if (outType == 0)  //user has entered -d
					 {
						outType = 2;
					 }
				  break;

				case 's':
				  PI_Prob = atof(argv[i]);
				  break;

				case 't':
				  PI_Density = atof(argv[i]);
				  break;
				  
				case 'l':
				  lut_size = atoi(argv[i]);
				  break;

				case 'r':
				  risefall_time_percentage = atof(argv[i]);
				  break;
   

         default:
				  printf ("Invalid command!\n Program aborted!\n\n");
                exit (1);
                break;
            }
        }
	if (i >=argc)
	   break;
        command = 0;
        
        if (argv[i][0] == '-')
        {
            command = 1;
            option = argv[i][1];
        }
        i++;
    }
    
    return 0;
}

/***************************************************************************/
/* Get the input name of a latch */
void getLatchInputName (char *signal_name,
								int latch_index)
{
  strncpy (signal_name, 
			  latch_in+latches[latch_index].inNameStart[0],
			  (latches[latch_index].inNameStart[1] 
				- latches[latch_index].inNameStart[0]));
  
  signal_name[latches[latch_index].inNameStart[1] 
				  - latches[latch_index].inNameStart[0]]= '\0';

}
/***************************************************************************/
/* Get the output name of a latch */
void getLatchOutputName (char *signal_name,
								 int latch_index)
{
  strncpy (signal_name, 
			  latch_out+latches[latch_index].outNameStart[0],
			  (latches[latch_index].outNameStart[1]
				- latches[latch_index].outNameStart[0]));
  signal_name[latches[latch_index].outNameStart[1] 
				 - latches[latch_index].outNameStart[0]] = '\0';

}
/***************************************************************************/
/* Get the input name of a LUT */
void getLUT_InputName (char *signal_name,
							  int LUT_index,
							  int input_index)
{
 strncpy (signal_name, 
			 inputNames+luts[LUT_index].inNameStart[input_index],
			 (luts[LUT_index].inNameStart[input_index+1] 
			  - luts[LUT_index].inNameStart[input_index]));

 signal_name[luts[LUT_index].inNameStart[input_index+1] 
				- luts[LUT_index].inNameStart[input_index]] = '\0';

}
/***************************************************************************/
/* Get the input name of a LUT */
void getLUT_OutputName (char *signal_name,
								int LUT_index)
{
  strncpy (signal_name, 
			  outputNames+luts[LUT_index].outNameStart[0],
			  (luts[LUT_index].outNameStart[1] 
				- luts[LUT_index].outNameStart[0]));
  signal_name[luts[LUT_index].outNameStart[1] 
				 - luts[LUT_index].outNameStart[0]]='\0';
}

/***************************************************************************/
/*  Find all feedbackloops in the circuit */
int find_FB_loop (int LUT_index,
						int* input_index)
{
  int input_count;
  int lut_count;
  int latch_count;  
  char LUT_inputName[80];
  char LUT_outputName[80];
  char latch_inputName[80];
  char latch_outputName[80];


  /* if the output of a latch is the input of a LUT, then a feedback loop
     exist. */

   for (latch_count = 0; latch_count<numLatches; latch_count++)
	 {
		getLatchOutputName(latch_outputName, latch_count); 

		for (input_count=0; 
			  input_count<luts[LUT_index].numInputs; 
			  input_count++)
		  {
			 getLUT_InputName(LUT_inputName, 
									LUT_index,
									input_count);
			 
			 if (strcmp (latch_outputName, LUT_inputName) == 0)
				{
				  (*input_index) = input_count;
				  return (1); // Yes, find a feedbackloop
				}
		
		  } // end for...input_count
		
	 } // end for...latch_count
  
  return (0); // no feedbackloop is found
}


/***************************************************************************/
/* Filter Mechanism */
/* Filter Out the short pulses when the transition density is higher than 
	half of the clock (when the density is higher than 1)
*/
int filter (double *probability,
				 double *density)
{
  double function0, function1;
  double filtered_prob, filtered_density;
  double filter_factor;
  int filtered;

  filtered = 0;
  if ((*density) > 1)
	 {

		if ((*density) < 20)
		  {
			 /* Transition Filter Mechanism is used to adjust the short pulses */
			 function0 = 1 - exp (-risefall_time_percentage*(*density)/(2*(1- (*probability))));
			 function1 = 1 - exp (-risefall_time_percentage*(*density)/(2*(*probability)));		  
			 filter_factor = (1-function0)*(1-function1)/(1-function0*function1);
			 filtered_density = (*density) * filter_factor;
			 filtered_prob = filter_factor* (1/(1-function0) - (*probability));
		   }
		else
		  {
			 /* 
				 Special Case:
				 If Transition Density is higher than 20, which is ten times 
				 faster than the clock frequency, we assume that all the short
				 pulse shouldn't pass through. As a result,
				 Transition Density = 2*(Probability)*(1-Probability)
			 */	 
			 filtered_prob = (*probability);
			 filtered_density = 2*(*probability)*(1-(*probability));
		  }

		(*probability) = filtered_prob;
		(*density) = filtered_density;

		filtered = 1;
	 }

  return(filtered);
}


/***************************************************************************/
/* Debugging Functions */

void printLutOrder()
{
  int i;
  // Print out the order of the luts to be calculated.
  printf ("\n");
  for (i = 0 ; i<numLuts; i++)
  {
    printf ("Order of luts to be done %d is lut %d\n",i,order[i]);
  }   
   
}

/***************************************************************************/
void printLutList()
{

  char input [80];
  char output [80];
  int i,j,l, k;;
  int function_number;

  function_number = (int) pow(2,lut_size); 

  for (i = 0 ; i < numLuts; i++)
  {
    printf ("\n\nLut %d\n", i);
    printf ("Number of inputs: %d\n", luts[i].numInputs);

    for (j = 0 ; j < luts[i].numInputs; j++)
    {
      strncpy (input, inputNames+luts[i].inNameStart[j],
	       luts[i].inNameStart[j+1]-luts[i].inNameStart[j] );
      input[ luts[i].inNameStart[j+1]-luts[i].inNameStart[j] ] = '\0';
            
      printf ("\nInput: %s\n", input);    
      printf ("   Probability input is high: %f\n", getProb(input));      
      printf ("   Probability value changes: %f\n" ,luts[i].inChange[j]);      
      printf ("   Transition Density:        %f\n",luts[i].inDensity[j]);
  
    }

    printf ("\n\n-- Output --\n");
    strncpy (output, outputNames+luts[i].outNameStart[0], 
            luts[i].outNameStart[1]-luts[i].outNameStart[0] ); 
    output[ luts[i].outNameStart[1]-luts[i].outNameStart[0] ] = '\0'; 
    
    printf ("Output: %s\n", output);
    printf ("   Probability output is high: %f\n", luts[i].outProb[1]);
    printf ("   Probability value changes:  %f\n", luts[i].outChange);
    printf ("   Transition Density:         %f\n", luts[i].outDensity);   


    // Print out the truth table of the LUTS.
 
    printf ("\nOutput is high for values: ");
    l =0;
    k= 0;

    while ((!l) && k <  function_number)
    {
      if (luts[i].outputHigh[k] == 1)
      {
	printf ("%d ",k);
      }
      else if (luts[i].outputHigh[k] == -1) 
      {
	l = 1;
      }
      k++;
    }

  }

  
}

/***************************************************************************/
void printArray(int inputArray[],
		int arraySize)
{
  int i;

   printf ("\n------ SRAM Table --------\n");
   for (i=0; i<arraySize; i++)
   {
     printf ("  %d\n",inputArray[i]);
   }
}

/***************************************************************************/
void printDouble(double doubleArray[],
		 int arraySize)
{
  int i;


   for (i=0; i<arraySize; i++)
   {
     printf ("  %f\n",doubleArray[i]);
   }
}




/***************************************************************************/

int main (int argc, char *argv[])
{
  int i;
  int times = 10;
  double error = 0.05;
  char *inFile ;
  char *outFile;
  char *netFile;
  int outType;
  char teststr[20];

  /* set default values */
  inFile = new char [80];
  inFile[0] = '\0';
  outFile= new char [80];
  strcpy (outFile, "activity.txt");
  netFile= new char [80];
  strcpy (netFile, "activity.act");
  outType = 2;
  PI_Prob = DEFAULT_PI_Prob;
  PI_Density = DEFAULT_PI_Density;
  lut_size = K_INPUT;
  risefall_time_percentage = DEFAULT_RISEFALL_TIME_PERCENTAGE;

  /* Step 1: Obtain input parameters and build the function table for 
     a particular LUT size */
  parseArgs (argc, argv, times, error, inFile, outFile, netFile, outType);
  
  build_table(); 


  /* Step 2: Extract data from .blif file */
  FILE *file;
  // cristinel.ababei:  I commented out the use of fpos_t
  // because was giving segmentation fault during execution.
  // fpos_t pos = 0; // set pos to the beginning of a file
 

  if (inFile[0] == '\0')
    {
      printf ("Missing filename object!\n");
      usage (argv[0]);
      exit(1);
    }
  
  if(!(file = fopen(inFile, "r")))
    {
      fprintf(stderr, "Unable to open %s for reading.\n\n", inFile);
      exit(2);
    }
    
  /* Allocate Memory for arrays to record information of LUTs and Latches */
  printf ("Allocate data.\n");
  readBlif(file, 0);  

    
  // reposition the pointer to the beginning of the file
  // cristinel.ababei:  I replaced next line with closing and opening
  // the file again.
  //  fsetpos (file, &pos);
  fclose(file);
  file = fopen(inFile, "r");

  // Record LUT and Latch information in .blif file into arrays
  printf ("Populate data structure.\n");
  readBlif (file, 1); 
  
  fclose(file);
  
    
  /* Step 3: Organize LUTs for Transition Density Calculation  */
  printf ("\nOrder Luts.\n");
  orderLuts();   
 
    printf ("Calculating static probablity\n");

	 /* Calculate probabilities for LUTs and Latches */

	 i =1;
    while( (calcProb(error) == 0) && (i<times) )
    {
        printf ("\nOne iteration completed.\n");
        i++;
    }

	 printf ("\nNumber of Iterations: %d\n\n",i);

    printf ("Calculating transition probability\n");
    calcChange();
    
    printf ("Calculating transition density\n");
    calcDensity();


	 /* Step 4: Record calculated probabilities and densities as files*/
    dump2File(outFile);
    
    dumpNetList(netFile, outType);
    
	 /* Step 5: Clean up */
	 freeMem();
	 printf ("\nProgram Ended\n");

	 delete_table();
    return 0;
}
