#ifndef PROB_CALC_H
#define PROB_CALC_H

#include <map>        
#include <string>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*-----------------------------------------------------------------------------
 *  Copyright 2002 By Kara K.W. Poon and Steve Wilton
 *
 * Permission to use, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 * Permission is granted ONLY for non-commercial use.  For
 * commercial use, please contact the authors.
 *
 * This software may be distributed (but not offered for sale or transferred
 * for compensation) to third parties, provided such third parties agree to
 * abide by the terms and conditions of this notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHORS, AS
 * WELL AS THE UNIVERSITY OF BRITISH COLUMBIA AND THE
 * UNIVERSITY OF TORONTO DISCLAIM ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL THE
 * AUTHORS OR THE UNIVERSITY OF BRITISH COLUMBIA OR THE
 * UNIVERSITY OF TORONTO BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *---------------------------------------------------------------------------*/

/* 
	
	Project Topic: Flexible Power Model for Field Programmable Gate Arrays
	Institute:     The University of British Columbia, Canada
	Supervisor:    Steve J. E. Wilton

	Program Name:  probcalc.h and probcalc.cpp
	Function:      This is an activity generator for estimating the switching
                  activity at the output of each LUT. 

	Input file:    .blif file from SIS
	Output file:   .txt file (describes the probabilities and transition 
	                          densities of inputs and the ouput)
						.act file (records the probabilities and transition 
                             densities. This file will be used as an input
                             file for TV-pack for power estimation purpose) 			  
									  
	History:
	July, 2001     Program was first developed by Andy Yan

	February, 2002 Modified by Kara Poon to handle different LUT sizes and 
                  take the static probabiliy and transition density for 
						the primary inputs as input variables to this program.  

	April, 2002    Modified by Kara Poon to include the Filter mechanism
                  of the transition density model					
                
 */

using namespace std;

/*  Constants */
#define MAXSIZE 1048576 // 1 MegaByte buffer
#define DEFAULT_PI_Prob 0.50  // Starting input probability that the signal is high 
#define DEFAULT_PI_Density 0.50
#define NUMBER_OF_INTERNAL_NODES 8
#define K_INPUT 4
#define NUMBER_OF_SRAMS_PER_LUT 16
#define INITIAL_SRAM_DENSITY 0
#define DEFAULT_RISEFALL_TIME_PERCENTAGE 0.1 
#define CLOCK_DENSITY 2

/* Create type definition for hash table 
   that maps a double with a string  */
typedef map <string, double> STRING2F;  

/* Number of Luts */
int numLuts ; 

/* Number of Latches */
int numLatches;

/* global variables for primary input probability and density */
float PI_Prob;
float PI_Density;
float Initial_FB_Density;

typedef struct tagLutData
{
   int numInputs;        /*  Number of inputs per Lut  */
   double *outProb;    /*  Probablity that the output value is high */
   double *inChange;   /*  Transition Probability that the input values changes
                             inchange is calculated as:
                             delta = 2p(1-p)   
									  where p is the probablity that the value is high
                         */
   double outChange;     /*  Transition Probability that the output value changes */
   double *inDensity;  /*  Transition density for the input values  */ 
   double outDensity;    /*  Transition density for the output values */

   int *inNameStart ;  /*  5-value Array pointer.  This data structure 
									  points to the starting and ending names for the 
									  input signals. */ 
   
   int *outNameStart ; /*  Data structure that points to the starting and 
									  ending names for the output name.  */ 

   int *outputHigh ;  /*  Output values that are high corresponding 
									  to bit inputs */

} LUTDATA;

LUTDATA * luts ;           /*  Declare the number of luts[numLuts] */

char inputNames[MAXSIZE];  /*  Buffer that stores the input names 
										 of all the luts  */
char outputNames[MAXSIZE]; /*  Buffer that stores the output names 
										 of all the luts */

int *linked ;              /* Flag variable that determines if the 
										a particular lut has been ordered yet */
int * order ;              /* order[numLuts].  Order to do array. */



typedef struct tagLatchData
{
  double inProb;         /*  Probablity that the latch input is high  */
  double outProb;        /*  Probablity that the latch output is high */
  double outDensity;     /*  Transition Density at the output */

  int inNameStart[2];    /*  Data structure that points to the 
										starting and ending input name  */
  int outNameStart[2];   /*  Data structure that points to the 
									  starting and ending output name */
} LATCHDATA;

LATCHDATA * latches;

char latch_in [MAXSIZE];  /*  Buffer that stores the input names 
										of all the latches */
char latch_out [MAXSIZE]; /*  Buffer that stores the output names 
										of all the latches */

STRING2F netList;         /*  Hash table that maps signal names with 
										probability values */

char** table;             /*  Truth table for the LUT functions */
int lut_size;             /*  Number of input per look-up-table */
float risefall_time_percentage; /* Rise and fall time value relative
                                   to the clock period (for filter
                                   mechanism */  

#endif //PROB_CALC_H   
