#include <stdio.h>
#include <string.h>
#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "graphics.h"
#include "read_netlist.h"
#include "read_activity.h"
#include "print_netlist.h"
#include "read_arch.h"
#include "draw.h"
#include "place_and_route.h"
#include "stats.h"
#include "path_delay.h"

/*----------------------------------------------------------------------------
 *  Copyright 2002 By Kara K.W. Poon and Steve Wilton
 *
 * Permission to use, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 * Permission is granted ONLY for non-commercial use.  For
 * commercial use, please contact the authors.
 *
 * This software may be distributed (but not offered for sale or transferred
 * for compensation) to third parties, provided such third parties agree to
 * abide by the terms and conditions of this notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHORS, AS
 * WELL AS THE UNIVERSITY OF BRITISH COLUMBIA AND THE
 * UNIVERSITY OF TORONTO DISCLAIM ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL THE
 * AUTHORS OR THE UNIVERSITY OF BRITISH COLUMBIA OR THE
 * UNIVERSITY OF TORONTO BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *----------------------------------------------------------------------------*/
/*
  VPR was first written by Vaugh Betz at the University of 
  Toronto. Descriptions on different versions of VPR is available
  in the descript.txt file.

  The power model is an extension of the original VPR. This 
  feature is added by Kara Poon at the University of British 
  Columbia as a master thesis project supervised by Dr. Steve
  Wilton.

 */

/***************************************************************************/
/*
  Project: Power Estimation for FPGAs
  Institute: The University of British Columbia
  Developer: Kara Poon
  Supervisor: Dr. Steve Wilton

  This source file serves as an add-on feature to the original VPR program. 
  It compute the dynamic, short-circuit, and leakage power dissipation. 

  Dynamic Power is calculated based on the activity information from
  the activity estimator.

  Short-Circuit depends on the "short_circuit_power_percentage" value
  in the architecture file. The short-circuit power is modelled as a 
  percentage of the dynamic power. The percentage is a user-defined value.

  Leakage power Assumptions:
    1. Transistors inside the unused switches and CLBs are assumed to 
      dissipate leakage power
    2. Transistors inside the used switches and CLBs dissipate leakage
      power during the non-transition time.
    3. Leakage current of the SRAM cell depends on the "SRAM_leakage" value
      in the architecture file. This is a user-defined value.	 
    4. Leakage power of the clock network is not considered  	 

  Note: Transition time is the time when the node in the circuit is charging
        and discharging (which is approximately the rise and fall time)

	Non-transition time is the time when the node is inactive
	For example: if the transition time for a node is 0.1ns, and its
	switching frequency is 100Mhz (10ns per cycle). Then
	the non-transition time for this node is 0.9ns.
	We assume the transition time is very small compared to the
        non-transition time.					

  Four output files will be produced:
  ROUTING_power.echo : contains routing power dissipation for each net
  LBpower.echo       : contains logic block power dissipation for each LUT
  CLKpower.echo      : contains the clock power dissiaption 
  power.echo         : contains the summary of power dissipation for this
                       circuit

 */

// cristinel.ababei
// For leakage-power calculations (see rr_graph3.c)
extern int   counter_conn_buffers[2];
extern int   counter_conn_pass_tran[2];
extern int   total_hwrp_switch_connections;
extern int   total_efective_eliminated_due_to_hwrp_insertion;
extern int   Lcounter[3];
extern int   Tcounter[3];
extern int   crosscounter[3];
extern int   HVcounter[3];

/******************* local variables for this file *************************/
/* Time per clock cycle */
float T_clock_cycle;

/* capacitance variables for power calculation */
/*
  components commonly used inside each logic block
  INV1x_Cin, INV1x_Cout: input and output capacitance for an 1X inverter
  INV2x_Cin, INV2x_Cout: input and output capacitance for an 2X inverter
  INV4x_Cin, INV4x_Cout: input and output capacitance for an 4X inverter
   */
float INV1x_Cin, INV1x_Cout, INV2x_Cin, INV2x_Cout, INV4x_Cin, INV4x_Cout;

/* For minimium size transistors */
/*
  NMOS_C_gate, PMOS_C_gate: Gate Capacitance
  NMOS_C_drain, PMOS_C_drain: Drain Capacitance
  NMOS_C_source, PMOS_C_source: Source Capacitance

  Note: for small transitors, the drain and source capacitance are the
        same.
 */
float NMOS_C_gate, NMOS_C_drain, NMOS_C_source;
float PMOS_C_gate, PMOS_C_drain, PMOS_C_source;

/* leakage Power */
/* local variables for individual components within the architecture */
/* leakage variables for NMOS and PMOS minimum size transitors*/
float leakage_1X_nmos;
float leakage_1X_pmos;

/* leakage variables for 1X, 2X, 4X inverters */
float leakage_1X_inv;
float leakage_2X_inv;
float leakage_4x_inv;

/* leakage variable for clock buffer */
float leakage_clock_buffer;

/********************* Subroutines local to this module *****************/
// ======================================================================= //
static float find_routing_power (
    t_subblock_data *subblock_data_ptr,
    int num_switch,
    int num_segment,
    t_segment_inf *segment_inf,
    int total_segments_used,
    enum e_switch_block_type switch_block_type,
    float *routing_leakage,
    FILE *power_file,
    int best_channel_width );
/*
  Function: Calculate the routing power for each net
  output:   - routing power (dynamic and short circuit power)
            - routing leakage power 

  variable:
  *subblock_data_ptr: Pointer to a subblock data record
  num_switch: number of different type of switches
  num_segment: number of different type of segments
  *segment_inf: Pointer to a segment record
  total_segments_used: total number of segment used
  switch_block_type: switch block type (subset, Universal, Wilton, Wilton3) 
  *routing_leakage:   pointer to leakage power from the routing 
                     (including SRAMs and unused switches)
  *power_file:        output file name for routing power outputs
						
 */

/*logic block power is not going to be affected by introducing hard wired routing patterns*/
static float find_logic_block_power (
    t_subblock_data *subblock_data_ptr,
    float *logic_block_leakage,
    FILE *power_file
);
/*
  Function: Calculate the Logic Block power dissipation
  output:   - the logic block power (dynamic and short-circuit)
            - leakage power in logic block
 
  variable:
  *subblock_data_ptr: Pointer to a subblock data record
  *logic_block_leakage: Pointer to the leakage power of the logic block
  *power_file:        output file name for logic block power outputs 

 */

float power_per_LUT (
    int lut_size,
    int max_number_of_internal_MUX_per_input,
    float inputProb[],
    float inputDensity[],
    int   inputs[],
    char* function,
    float *outProb,
    float *outDensity,
    float *logic_block_leakage,
    FILE *power_file
);
/*
  Function: Calculate the power dissipated by each LUT
  output:   - Power dissipation inside a LUT (dynamic and short-circuit)
            - accumulated leakage power so far
				- output probability of the LUT
				- output denstiy of the LUT

  variable: 
  lut_size: look-up-table size
  max_number_of_internal_MUX_per_input: number of 2-input MUXs inside
  inputProb: Array for input probability
  inputDensity: Array for input density
  inputs:   Array for the indexes of input pins
  function: function implemented in the LUT
  outProb:  output probability of the LUT
  outDensity: output Density of the LUT
  logic_block_leakage: accumulated leakage power
  power_file:  output file name for the logic block power outputs
*/

void find_prob_and_density_for_MUX (
    float probA,
    float probB,
    float probC,
    float densityA,
    float densityB,
    float densityC,
    float *outProb,
    float *outDensity
);
/*
  Function: Find the probability and transition density for a 2-input MUX
  output:   - static probability at the output
            - transition density at the output 

  variable:
  Note: A = the first input of the MUX
        B = the second input of the MUX
		  C = the select signal of the MUX

  probA: static probability of the first input 
  probB: static probability of the second input 
  probC: static probability of the select input 
  densityA: transition density of the first input
  densityB: transition density of the second input 
  densityC: transition density of the select input 
  outProb:  static probability at the output
  outDensity: transition density at the output
  
 */

float SRAM_Value_To_float (
    char SRAM_value
);
/*
  Function: convert the SRAM_value to floating point
  output:   the floating point value for the SRAM

  variable:
  SRAM_value: a character represent a logic value in a SRAM

 */

/* Functions for calculating capacitance */
void calculate_individual_cap (void);
/*
  Function: Calculate the capacitance for componesnts that are commonly
            used within the logic block

 */

float calculate_gate_cap (
    float COX,
    float length,
    float width
);
/*
  Function: Calculate the gate capacitance
  output:   Gate Capacitance

  variables:
  length: length of the transistor (in m)
  width:  width of the transitor (in m)
  Cpoly:  gate-poly capacitance per m  

 */


void find_drain_source_cap_small (
    float length,
    float width,
    float* C_drain,
    float* C_source,
    float CJ,
    float CJSW,
    float CJSWG,
    float CGDO
);
/*
	Function: Calculate the drain_source capacitance
	          for transistors smaller than the 
				 size limit
	Output:   Drain and source capacitance

	variables:
	length: length of the transistor (in um)
	width:  width of the transitor (in m)
	C_drain: Drain capacitance
	C_source: Source capacitance
	CJ:       Junction capacitance
	CJSW:     side-wall junction capacitance
	CJSWG:    gate-edge side-wall bulk junction capacitance
	CGDO:     gate-drain overlap capacitance
 */

void find_drain_source_cap_large (
    float length,
    float width,
    float* C_drain,
    float* C_source,
    float CJ,
    float CJSW,
    float CJSWG,
    float CGDO
);
/*
	Function: Calculate the drain_source capacitance
	          for transistors larger than the 
				 size limit
	Output:   Drain and source capacitance

	variables:
	length: length of the transistor (in m)
	width:  width of the transitor (in m)
	C_drain: Drain capacitance
	C_source: Source capacitance
	CJ:       Junction capacitance
	CJSW:     side-wall junction capacitance
	CJSWG:    gate-edge side-wall bulk junction capacitance
	CGDO:     gate-drain overlap capacitance
 */


void find_cap (
    tx_type transistor_type,
    float length,
    float width,
    float* C_drain,
    float* C_source,
    float* C_gate
);

/*
  Function : Find the capacitance of a transistor
  output: Drain, source, and gate capacitance
  variables:
        length: channel length of the transistor
        width: channel width of the transistor
        C_drain: Drain capacitance
        C_source: Source capacitance
        Cgate: Gate capaciatance
*/      

void calculate_INV_cap (
    float PMOS_length,
    float PMOS_width,
    float NMOS_length,
    float NMOS_width,
    float *Cin,
    float *Cout
);

/*
  Function : Calculate the input and output capacitance of an inverter
  output: input and output capacitance of the inverter
  variables:
        PMOS_length: transistor length of the PMOS
        PMOS_width:  transistor width of the PMOS
        NMOS_length: transistor length of the NMOS
        NMOS_width:  transistor length of the NMOS
        *Cin: pointer to input capacitance of the inverter
        *Cout: pointer to output capacitance of the inverter
*/



float calculate_DFF_power (
    float inputProb,
    float inputDensity,
    float clockProb,
    float clockDensity,
    float *outProb,
    float *outDensity,
    FILE *power_file
);
/*
  Function : Calculate the power of a D Flip Flop
  output:    - probability and transition density at the output
  
  variables:
  inputProb: input probability
  inputDensity: input transition density
  clockProb: probability of the clock signal
  clockDensity: transition density of the clock signal
  *outProb: Pointer to the output probability
  *outDensity: Pointer to the output density
  *power_file: output file
*/

float power_per_subblock (
    int clock_input,
    int lut_size,
    int max_subblock_per_block,
    float internal_Cwire,
    float input_and_output_probability[],
    float input_and_output_density[],
    t_subblock subblock_ptr,
    float *logic_block_leakage,
    FILE *power_file
);
/*
  Function : Calculate power dissipated inside the subblock
  output:    - return the dynamic and short circuit power of 
               the subblock
             - leakage power
  variables: 
  clock_input: number of clock inputs (default is 1)
  lut_size:    look-up-table size
  max_subblock_per_block: maximum number of subblocks per
                          logic block
  internal_Cwire: capacitance of the internal wire
  input_and_output_probability: array contains the probability
                                of the inputs and outputs
										  
  input_and_output_density: array contains the transition
                            density of inputs and outputs
  subblock_ptr: pointer to a subblock
  *logic_block_leakage: pointer to the logic block leakage 
                        variable
  *power_file: output file 	 							
 */


float power_per_input_MUX (
    int overall_input_to_MUX,
    int inputnum,
    float inputProb[],
    float inputDensity[],
    int level,
    char  SRAM_values[],
    float *logic_block_leakage,
    float *outProb,
    float *outDensity
);

/*Satish. When the number of tracks increases for hardwired routing patterns, the number of muxes in the connection boxes will go up. This could be one reason for the increase in overall consumption with hwrp though we expect a decrease in power consumption as a result of decreasing some of the pass transistors in the switch boxes*/

/*
  Function : calculate the power dissipation for a MUX
  output:    - dynamic and short-circuit power dissipated in the
               MUX
				 - accumulated value for leakage power	
             - probability and transition density at the output

  variables:
  overall_input_to_MUX: maximum number of inputs to the MUX
  inputnum: number of inputs used
  inputProb: Array for input probabilities
  inputDensity: array for input transition density
  level: number of columns of 2-input MUXs in the big Mux
  SRAM_values: Array of SRAM values stored for the output selects
  *logic_block_leakage: pointer to the logic block leakage 
                        variable
  *outProb: pointer to output probability variable
  *outDensity: pointer to output transition density variable
 */


void integer_to_SRAMvalues (
    int SRAM_num,
    int input_integer,
    char SRAM_values[]
);
/*
  Function : Change the selected integer to a stream of binary value
  output:    a stream of binary value for the SRAM cells
  
  variables:
  SRAM_num: length of the binary stream
  input_integer: an integer represents a selected output index 
                 for the MUX
  SRAM_values: character array for the binary bit stream
*/

void find_SRAMvalues_for_input_MUX (
    int SRAM_num,
    int input_clb_pin_number,
    int max_subblock_per_block,
    int clock_input,
    char SRAM_values[]
);
/*
  Function : determine the binary values for the SRAM cells based
             on the selected input pin number
  output:    SRAM_cell bit stream
  variables:
  SRAM_num: length of the binary stream
  input_clb_pin_number: an integer represents a selected output index 
                 for the MUX
  SRAM_values: character array for the binary bit stream  

*/


int binary_str_to_integer (
    int binary_length,
    char* binary_str
);
/*
  Function : convert a binary bit stream back to an integer
  output:    an integer 
  variables:
  binary_length : length of the bit stream
  binary_str: character array contain the binary stream

 */


void reverse_binary_str (
    int binary_length,
    char* binary_str,
    char* reverse_binary_str
);
/*
  Function : Reverse a binary string
  output:    the reversed binary string
  variables:
  binary_length : length of the bit stream
  binary_str: character array contain the binary stream
  reverse_binary_str: the reversed binary string
 */

float find_clock_power (
    float *clock_leakage,
    FILE *power_file
);
/*
  Function : Calculate the power dissipation of the clock
  output:    - dynamic and short circuit power
             - clock leakage power
  variables:
  *clock_leakage: clock leakage power
  *power_file: output file for power output

*/

float find_clock_power_per_network (
    float *clock_leakage,
    int clock_index,
    int* total_clock_buffers,
    int* total_clock_segments,
    FILE *power_file
);
/*
  Function : Calculate the power dissipation per clock network
  output:    -Power dissipated (dynamic and short-circuit)
             - leakage power (if the network is not used )  

  variables:
  clock_leakage: leakage power
  clock_index: index for the clock network
  total_clock_buffers: total number of clock buffers in the 
                       network							  
  total_clock_segments: total number of clock segments in
                         the network
  *power_file: output file						 

 */

static boolean is_global_clock (
    int inet,
    int *num_subblocks_per_block,
    t_subblock **subblock_inf,
    float *clock_density
);
/*
  Function : Check if the net is a global clock network
  output:    return 1 if true; otherwise, return 0

  variables:
  inet: index of the net
  num_subblocks_per_block: number of subblocks per logic 
                           block
									
  subblock_inf: Record for subblock information
  clock_density: transition density of the clock
 */
// ======================================================================= //

/*Satish. I am not too interested in finding out the power in the clock network at this point since we have not modified any portion of the clock network*/
void find_clock_power_per_level (
    int clock_index,
    int current_level,
    int length_of_segment_at_this_level,
    float buffer_delay,
    float segment_delay,
    float buffer_needed_per_segment_length,
    float *total_delay_at_this_level,
    float *total_power_at_this_level,
    float *clock_leakage,
    int *total_clock_buffers,
    int *total_clock_segments
);
/*
  Function : Calculate the power dissipation at each level of 
             the clock network
  output:    - power dissipated at this level (when active)
             - leakage power (when the network is not active)
  
  variables:
  clock_index: index of the clock network
  current_level: current clock level
  length_of_segment_at_this_level: length of the segment at 
                                   this level
  buffer_delay: single buffer delay
  segment_delay: unit segment delay
  buffer_needed_per_segment_length: additional buffer per unit 
                                    segment
  *total_delay_at_this_level: pointer to a  value for delay
  *total_power_at_this_level: pointer to a value for power
  *clock_leakage: clock leakage power
  *total_clock_buffers: pointer to accumulated value for buffers
  total_clock_segments: pointer to accumulated value for segments

 */

// ======================================================================= //

/* leakage power calculation */
void calculate_individual_leakage (
    int num_switch
);
/*
  Function : Calculate the leakge power for all basic components:
             Pass transistors, inverters, SRAM cells, and
				 switches
   
  variables:
  num_switch: number of switch types

 */
// ======================================================================= //

void calculate_unused_CLB_leakage (
    int maximum_subblocks_per_block,
    int lut_size,
    int clock_input,
    float *logic_block_leakage
);
/*
  Function : Calculate the leakage power for unused CLBs

  variables:
  maximum_subblocks_per_block: maximum number of subblocks(LUTs) in
                               a logic block
  lut_size: LUT input size
  clock_input: number of clock input (DEFAULT:1)
  *logic_block_leakage: pointer to logic block leakage power

 */
// ======================================================================= //

float calculate_leakage_power (
    tx_type transistor_type,
    float diffusion_cap,
    int size
);
/*
  Function : Calculate the subthreshold current for a transistor
  output:    - subthreshold current for the transistor

  variables:
  transistor_type: NMOS or PMOS
  diffusion_cap:   diffusion capacitance (in F)
  size:            size of the transitor
 */
// ======================================================================= //

float get_NFS (
    tx_type transistor_type,
    int temperature
);
/*
  Function : Get the NFS value for the operation temperature
  output:    NFS value

  variables:
  transistor_type: NMOS or PMOS
  temperature: operation temperature in Degree Celcius
 */
// ======================================================================= //

//Satish. This is the one i have to concentrate on
float calculate_switch_leakage(
			       boolean buffered,
			       float Cout,
			       int iswitch); // added by cristinel.
/*
  Function : Calculate the leakage power of the switch

  variables:
  buffered: yes = the switch is a tri-state buffer 
            no  = the switch is a pass transistor 
  Cout:     output capacitance of the switch 	
  iswitch:   index of switch type.		
 */
// ======================================================================= //

float calculate_INV_leakage (
    int inverter_size,
    float diffusion_cap
);
/*
  Function : Calculate the leakage power for a inverter 

  variables:
  inverter_size: size of the inverter
  diffusion_cap: output diffusion capacitance for
                 the inverter
  
 */
// ======================================================================= //


/***************************************************************************/
/* Calculate the power dissipation for the circuit */
float get_power_per_clock_cycle (
    t_subblock_data *subblock_data_ptr,
    int num_switch,
    float T_crit,
    int num_segment,
    t_segment_inf *segment_inf,
    int total_segment_used,
    enum e_switch_block_type switch_block_type,
    int best_channel_width )
{
    FILE *fp;
    float routing_energy, logic_block_energy, clock_energy, total_energy;
    float routing_power, logic_block_power, clock_power, misc_energy;
    float routing_leakage, logic_block_leakage, clock_leakage;
    float total_power_dissipation, total_leakage;
    float routing_power_percentage, logic_block_percentage, clock_power_percentage;
    float routing_leakage_percentage, logic_block_leakage_percentage, clock_leakage_percentage;
    float total_leakage_percentage;

    T_clock_cycle = T_crit;

    calculate_individual_cap();
    calculate_individual_leakage(num_switch);
    //printf("\n Sram leakage is %f\n",SRAM_leakage);

    /* Calculate energy in parts */

    /* First, calculate energy due to routing channels */
    fp = my_fopen ("ROUTING_power.echo", "w", 0);
    
    //Satish. Looks like routing power will include the switch boxes too. so it is going to be more diff to isolate which component of power is going up in case of hwrp
    routing_power = find_routing_power(subblock_data_ptr, num_switch,
                                       num_segment, segment_inf,
                                       total_segment_used,
                                       switch_block_type,
                                       &routing_leakage,  // Will be updated in!
				       fp,
				       best_channel_width);

    // cristinel: routing_leakage should include leakage at the level of
    // connections inside switch-blocks?
    routing_power += routing_leakage; // Include leakage power too.
    fclose (fp);

    /* Now, calculate energy due to logic blocks */
    fp = my_fopen ("LBpower.echo", "w", 0);
    logic_block_leakage = 0;
    logic_block_power = find_logic_block_power(subblock_data_ptr,
                        &logic_block_leakage,             // Will be updated in!
                        fp);
    logic_block_power += logic_block_leakage; // Include leakage power too.
    fclose (fp);

    /* Now, calculate energy due to clock distribution network */
    fp = my_fopen ("CLKpower.echo", "w", 0);
    clock_leakage = 0.0;
    clock_power = find_clock_power(&clock_leakage,fp);// Will be updated in!

    clock_power += clock_leakage;// It's zero allways because it's not accounted for.
    fclose (fp);

    /* Now, calculate energy due to anything else */

    misc_energy = 0.0;



    /* Power Analysis */
    total_power_dissipation = 
      routing_power + logic_block_power + clock_power + misc_energy;

    routing_power_percentage = routing_power/total_power_dissipation * 100;
    logic_block_percentage = logic_block_power/total_power_dissipation * 100;
    clock_power_percentage =  clock_power/total_power_dissipation * 100;

    total_leakage = routing_leakage+logic_block_leakage+clock_leakage;
    total_leakage_percentage = total_leakage/total_power_dissipation * 100;

    routing_leakage_percentage = routing_leakage/total_power_dissipation*100;
    logic_block_leakage_percentage = logic_block_leakage/total_power_dissipation*100;
    clock_leakage_percentage = clock_leakage/total_power_dissipation*100;

    routing_energy = routing_power*T_clock_cycle;
    logic_block_energy = logic_block_power*T_clock_cycle;
    clock_energy = clock_power*T_clock_cycle;
    total_energy = total_power_dissipation*T_clock_cycle;


    /* write to power report file */
    fp = my_fopen ("power.echo", "w", 0);
    fprintf (fp, "Critical_Path: %g\n",T_clock_cycle);
    fprintf (fp, "layout of FPGA: %d x %d\n", nx, ny);
    fprintf (fp, "operation temperature: %d \n", operation_temperature);
    fprintf (fp, "inputs: %d, outputs: %d, total clbs: %d, total global: %d\n",
             num_p_inputs, num_p_outputs, num_clbs, num_globals);
    fprintf (fp, "net num: %d, block num: %d\n\n",
             num_nets, num_blocks);

    fprintf (fp, "Power Analysis:\n");
    fprintf (fp, "routing Power: %g W(%f percent)\n",
             routing_power, routing_power_percentage);
    fprintf (fp, "Total Logic Block Power: %g W(%f percent)\n",
             logic_block_power, logic_block_percentage);
    fprintf (fp, "Clock Power Dissipation: %g W(%f percent)\n",
             clock_power, clock_power_percentage);
    fprintf (fp, "-----------------------------------------\n");
    fprintf (fp, "Total Power Dissipation: %g W\n\n", total_power_dissipation );


    fprintf (fp, "Leakage Power Analysis:\n");
    fprintf (fp, "Routing Leakage Power: %g W (%f percent)\n",
             routing_leakage, routing_leakage_percentage);
    fprintf (fp, "Logic Block Leakage Power: %g W (%f percent)\n",
             logic_block_leakage, logic_block_leakage_percentage);
    fprintf (fp, "Clock Leakage Power: %g W (%f percent)\n",
             clock_leakage, clock_leakage_percentage);
    fprintf (fp, "-----------------------------------------\n");
    fprintf (fp, "Total Leakage Power: %g W \n(%f percent of the total power dissipation)\n\n",
             total_leakage, total_leakage_percentage);

    fprintf (fp, "Energy Analysis:\n");
    fprintf (fp, "Routing Energy: %g J\n", routing_energy);
    fprintf (fp, "Logic Block Energy: %g J \n", logic_block_energy);
    fprintf (fp, "Clock Energy: %g J \n", clock_energy);
    fprintf (fp, "-----------------------------------------\n");
    fprintf (fp, "Total Energy: %g J \n\n",  total_energy);

    fprintf(fp, "Power Analysis Completed\n");
    fclose (fp);
    /*write pattern report file*/
    fp = my_fopen ("pattern.echo", "w",0);
    fprintf(fp, "Pattern Length Analysis\n");
    fprintf(fp, "segment length 1 results...\n");
    fprintf(fp, "L = %d\n",Lcounter[0]);
    fprintf(fp, "T = %d\n",Tcounter[0]);
    fprintf(fp, "+ = %d\n",crosscounter[0]);
    fprintf(fp, "H,V = %d\n",HVcounter[0]);
    fprintf(fp, "segment length 2 results...\n");
    fprintf(fp, "L = %d\n",Lcounter[1]);
    fprintf(fp, "T = %d\n",Tcounter[1]);
    fprintf(fp, "+ = %d\n",crosscounter[1]);
    fprintf(fp, "H,V = %d\n",HVcounter[1]);
    fprintf(fp, "segment length 6 results...\n");
    fprintf(fp, "L = %d\n",Lcounter[2]);
    fprintf(fp, "T = %d\n",Tcounter[2]);
    fprintf(fp, "+ = %d\n",crosscounter[2]);
    fprintf(fp, "H,V = %d\n",HVcounter[2]);

    fclose (fp);


    printf ("Total Power Dissipation: %g W\n\n", total_power_dissipation );

    printf ("Power analysis results are provided in the following files:\n");
    printf ("  ROUTING_power.echo: contains routing power dissipation \n");
    printf ("  LBpower.echo:       contains logic block power dissipation \n");
    printf ("  CLKpower.echo:      contains clock power dissipation \n");
    printf ("  power.echo:         contains overall power analysis\n\n");

    return(total_power_dissipation);
}//eof
// ======================================================================= //



/*Calculate the routing power for each net*/
static float find_routing_power (
				 t_subblock_data *subblock_data_ptr,
				 int num_switch,
				 int num_segment,
				 t_segment_inf *segment_inf,
				 int total_segments_used,
				 enum e_switch_block_type switch_block_type,
				 float *routing_leakage, // Will be updated in!
				 FILE* power_file,
				 int best_channel_width )
{
  int inet, inode, switch_count, seg_count;
  struct s_trace *tptr;
  float total_routing_power;
  float net_energy;
  float leakage_power;
  float Cforthisnet;
  int clock_index;
  int total_available_segments;
  float  total_segment_utilization;
  int by_pass_switch_num;

  // --- Assume no leakage current if the node  when the node is switching
  float overestimate_leakage;

  // cristinel.ababei
  int switch_conn_inside_CB =0; // same for usual-VPR or VPR-HWRP
  int switch_conn_inside_SB =0;
  int unused_buffers_inside_SB =0;
  int unused_pass_tran_inside_SB =0;
  float leak_power_in_buffer =0;
  float leak_power_in_pass_tran =0; 


  fprintf (power_file, "***************** Routing Power Analysis ************\n");
  total_routing_power = 0.0;
  clock_index =0;
  total_segment_utilization =0;


    /* step through all nets in the design */
    // ----------------------------------------------------------------------
    for(inet=0;inet<num_nets;inet++) {

        if (is_global[inet]) {


            if (is_global_clock(inet,
                                subblock_data_ptr->num_subblocks_per_block,
                                subblock_data_ptr->subblock_inf,
                                &clock_inf[clock_index].clock_density))
            {
                clock_index++;
            }
            continue;
        }

        /* for this net, step through all tracks that make up this
	   net.  For each track, the capacitance at that track is 
	   available  in the rr_node structure */

        Cforthisnet = 0.0;
        tptr = trace_head[inet];
        while(tptr != NULL) {
            inode = tptr->index;
            Cforthisnet += rr_node[inode].C;

            /* Added by JL, 2002 */
            /* consider cap between cascaded buffer stages */
	    // cristinel.ababei:  Here ".Cinternal" will reflect
	    // the fact that this switch is hard-wired or not by not
	    // including capacitance of the pass transistor. See read_arch.c
	    // last function.
            Cforthisnet += switch_inf[tptr->iswitch].Cinternal;

            /* counting the number for each type switch being used */
            switch_inf[tptr->iswitch].num_used ++;
            tptr = tptr->next;
        }


        /* calculate energy for this net, and add it to the
	   total routing energy */
        net_energy =
	            0.5*swing_voltage*supply_voltage
                    *(1+short_circuit_power_percentage)
                    *Cforthisnet /T_clock_cycle*net[inet].density;

        fprintf (power_file, "net %d: %g W\n", inet, net_energy);

        total_routing_power += net_energy;
    } // for (inet=0;inet<num_nets;inet++)
    // ----------------------------------------------------------------------



    // ----------------------------------------------------------------------
    // cristinel.ababei:   Whole rest of this routine used to calculate
    // leakage power due to unused connection-switches (just as a side note:
    // I think it is erroneous anyway). I replace it with what's after it
    // to account for hwrp's stuff (if required)
    // --- 1 --- Next is original part of this code
    /*---
    // --- Calculate the number of segments in the nets
    for (seg_count =0; seg_count< num_segment; seg_count++)
      {
	total_segment_utilization += 
	  segment_inf[seg_count].utilization*segment_inf[seg_count].frequency;
      }
    total_available_segments = (int)total_segments_used/total_segment_utilization;


    if (switch_block_type == SUBSET||switch_block_type == WILTON3)
      {
        by_pass_switch_num = subset_by_pass_switch_num;
      }
    else
      {
        by_pass_switch_num = other_by_pass_switch_num;
      }


    // --- To be conservative, we assume all the unused switches in the design
    // is leaking

    // --- Calculate the number of unused switch in the nets 
    for (seg_count =0; seg_count< num_segment; seg_count++)
      {
	// --- Calculate the number of connection switches along each segment
	// These connections are inside Connection-Blocks (CB)
        switch_inf[segment_inf[seg_count].opin_switch].total_num += 
	  total_available_segments
	  * segment_inf[seg_count].frequency
	  * segment_inf[seg_count].frac_cb;

        // --- Calculate the number of wire switches along each segment 
        // Take into account the situation when a wire bypasses a segment
	// cristinel.ababei:  This has to be re-thought.
	// These connections are inside Switch-Blocks (SB)
        switch_inf[segment_inf[seg_count].wire_switch].total_num += 
	  total_available_segments
	  * segment_inf[seg_count].frequency
	  * ((segment_inf[seg_count].frac_sb
	      *(segment_inf[seg_count].length+1)-2)
	     * by_pass_switch_num) + 2*F_switch;
	
      }//for


    // --- Calculate overall leakage power for unused switches
    leakage_power = 0;
    for (switch_count=0; switch_count<num_switch-2; switch_count++)
      {
	//During the transition time, the used switches are active 
	//and do not dissipate any leakage power. However, the 
	//transition time is very small compare to the non-transition 
	//time. Therefore, the used switch are inactive most of the 
	//time. We assume all the switches are inactive and leaking 
	//first. Then, depend on the type of the used switch, we 
	//estimate the portion that have been overestimated and 
	//deduct the portion from the leakage power calculated. 
	if (switch_inf[switch_count].buffered)
	  {
	    //if the switch is a tri-state buffer then
	    //one of the transistor (NMOS or PMOS) of the
	    //inverters would be leaking during inactive time.
	    //Assume the leakage current through PMOS
	    //and NMOS transistors of the inverter is
	    //about the same.
	    overestimate_leakage = 0.5;
	  }
        else
	  {
            // if the switch is a pass-transistor, then
	    // it should not leak when chosen 
            overestimate_leakage = 1;
	  }

        // --- Total leakage power is the sum of the leakage power
	// from the SRAM cell, and the unused switches
	// cristinel.ababei:  switch-connections corresponding to hard-wired
	// -routing-patterns do not have pass transistor so they should not have
	// leakage of SRAM (which controls the pass-tran), which does not exit 
	// itself.
	if ( switch_inf[switch_count].is_hwrp == FALSE ) {
	  leakage_power +=
	    (switch_inf[switch_count].total_num * SRAM_leakage  // 1
	     + ( switch_inf[switch_count].total_num 
		 - (switch_inf[switch_count].num_used * overestimate_leakage) ) 
	     * switch_inf[switch_count].Leakage_Power); // 2:  This second component
	  // due to ".Leakage_Power" already reflects that it's hwrp or not 
	  // (i.e., includes leakage of the pass transistor or not). See
	  // calculate_switch_leakage() at the end of this file.
	}
	else {
	  leakage_power +=
	    ( 0 // 1: This component is zero now.
	      + ( switch_inf[switch_count].total_num 
		  - (switch_inf[switch_count].num_used * overestimate_leakage) ) 
	      * switch_inf[switch_count].Leakage_Power ); // 2
	}

    }//for
    ---*/
    // ----------------------------------------------------------------------



    // ----------------------------------------------------------------------
    /*---
    // cristinel.ababei:  Next is the new way of computing leakage due to
    // unused switch-connections. For now this does not include those
    // inside SB between channels x,y and IO channels, which are anyway 
    // always there either for usual-VPR or for VPR-HWRP
    // --- 2 --- Next is my 1st modification
    // --- Calculate the number of segments in the nets
    for (seg_count =0; seg_count< num_segment; seg_count++)
      {
	total_segment_utilization += 
	  segment_inf[seg_count].utilization*segment_inf[seg_count].frequency;
      }
    total_available_segments = (int)total_segments_used/total_segment_utilization;


    // --- To be conservative, we assume all the unused switches in the design
    // is leaking


    // Not needed anymore
    // --- Calculate the number of unused buffers and pass-trans (also gives
    // number of unused SRAMs)
    unused_buffers_inside_SB += 
      ( counter_conn_buffers[0] - switch_inf[0].num_used ) + // (i) usual
      ( counter_conn_buffers[1] - switch_inf[1].num_used );  // (ii) hwrp
    // oBs:  if disable_hwrp==true, then ,above, second term of summation
    // will automatically be zero.
    unused_pass_tran_inside_SB += 
      ( counter_conn_pass_tran[0] - switch_inf[0].num_used ); // (i) usual
    // oBs:  There is no unused pass-trans for hwrp's because they do not
    // use pass-tras at all. In the above summation counter_conn_pass_tran[0]
    // reflects anyway the fact that there are hwrp's inserted or 
    // not, by the way it is computed in generate_distributions_hwrp()
    // in rr_graph3.c
    //Satish. Even buffers which are used leak during the time, they are inactive (not in transition). Check if that is also accounted for and it is not only the unused buffers whose leakage is computed

    // --- Calculate number of unused switch-connections inside CB's
    // which is pretty much the same for both usual-VPR or VPR-HWRP. 
    for (seg_count =0; seg_count< num_segment; seg_count++)
      {
	// --- Calculate the number of connection switches along each segment
	// These connections are inside Connection-Blocks (CB)
	switch_conn_inside_CB += 
	  total_available_segments
	  * segment_inf[seg_count].frequency
	  * segment_inf[seg_count].frac_cb;
      }//for

    // --- Calculate overall leakage power for unused switch-connections
    leakage_power = 0;

    // --- Add leak power in unused buffers and unused pass-trans.

    // By the way ".Leakage_Power" already reflects that it's hwrp or not 
    // (i.e., includes leakage of the pass transistor or not. 
    // See calculate_switch_leakage() at the end of this file.), 
    // I can calculate the following two quantities:
    leak_power_in_buffer = switch_inf[1].Leakage_Power;
    leak_power_in_pass_tran = 
      switch_inf[0].Leakage_Power - switch_inf[1].Leakage_Power;

    leakage_power +=
      // leak power in remained buffers of (i) usual switch-connections.
      // I add "switch_conn_inside_CB" because "switch_inf[0].num_used"
      // includes also used switch-conns inside CBs
      ( counter_conn_buffers[0] + switch_conn_inside_CB - switch_inf[0].num_used ) 
      * leak_power_in_buffer
      +
      // leak power in remained pass-tran of (i) usual switch-connections
      ( counter_conn_pass_tran[0] + switch_conn_inside_CB - switch_inf[0].num_used) 
      * leak_power_in_pass_tran
      +
      // leak power in remained buffers of (ii) hwrp switch-connections
      ( counter_conn_buffers[1] - switch_inf[1].num_used )
      * switch_inf[1].Leakage_Power; // (ii) hwrp

    // --- Add leakage in all (?) SRAMs
    leakage_power +=
      ( switch_conn_inside_CB + counter_conn_pass_tran[0] ) * SRAM_leakage;
    ---*/
    // ----------------------------------------------------------------------


    // ----------------------------------------------------------------------
    // --- 3 --- Next is my 2nd modification. Hopefully better now.
    // cristinel.ababei: Ok, let's put things in order:
    // *) First of all, leakage due to reverse-biasing of D,S diodes is not considered
    // at all in the "powermodel" project itself.
    // *) Roughly speaking, an used switch-connection leaks inside its buffer,
    // and inside the SRAM controlling the pass-tran (at the out of the buffer)
    // It does not leak in the pass-tran because its Vds is theoretically zero
    // after the transition period is gone.
    // *) When a switch-connection is not used, then we have leak in: buffer,
    // SRAM, and pass-tran (if Vds across it is >0 and this is subject of how
    // connections are made and their voltages during static functioning). Also,
    // if the buffer is shared and used already by a used switch-connection
    // I am not sure if the leakage inside the buffer should added again (it is
    // already added to account for used switch-conn) even though the pass-tran
    // which is unused may leak and its leak-current will flow thru the out inverter
    // of this shared buffer... For the moment let's assume that we count all 
    // buff, sram, and pass-tran leaks for unused switch-conns with the risk of
    // overestimation (i.e., make the whole situation as if bufferes were not shared?).

    // --- Calculate the number of segments in the nets
    for (seg_count =0; seg_count< num_segment; seg_count++)
      {
	total_segment_utilization += 
	  segment_inf[seg_count].utilization*segment_inf[seg_count].frequency;
      }
    total_available_segments = (int)total_segments_used/total_segment_utilization;
    if (switch_block_type == SUBSET||switch_block_type == WILTON3)
      {
        by_pass_switch_num = subset_by_pass_switch_num;
      }
    else
      {
        by_pass_switch_num = other_by_pass_switch_num;
      }

    // --- Calculate number of unused switch-connections inside CB's
    // which is pretty much the same for both usual-VPR or VPR-HWRP. 
    for (seg_count =0; seg_count< num_segment; seg_count++)
      {
	// --- Calculate the number of connection switches along each segment
	// These connections are inside Connection-Blocks (CB)
	switch_conn_inside_CB += 
	  total_available_segments
	  * segment_inf[seg_count].frequency
	  * segment_inf[seg_count].frac_cb;
        // --- Calculate the number of wire switches along each segment 
	// These connections are inside Switch-Blocks (SB)
	switch_conn_inside_SB +=  // ??? Is it correct ???
	  total_available_segments
	  * segment_inf[seg_count].frequency
	  * ((segment_inf[seg_count].frac_sb
	      *(segment_inf[seg_count].length+1)-2)
	     * by_pass_switch_num) + 2*F_switch;
      }//for
    // I think the above calculation of "switch_conn_inside_CB" and 
    // "switch_conn_inside_SB" as in the original powermodel code is wrong.
    //printf ("\n switch_conn_inside_CB in original powermodel: %d",switch_conn_inside_CB);
    //printf ("\n switch_conn_inside_CB by my calculations: %d",
    //	    nx*nx*5*best_channel_width);
    //printf ("\n switch_conn_inside_SB in original powermodel: %d",switch_conn_inside_SB);
    //printf ("\n switch_conn_inside_SB by my calculations: %d",counter_conn_pass_tran[0]);
    // Therefore I make the correction here (oBs: with a certain error because
    // I do not count connections between tracks in IO channels - at the fpga's
    // perifery - and tracks in xy channels:
    switch_conn_inside_SB = counter_conn_pass_tran[0];
    switch_conn_inside_CB = nx*nx*5*best_channel_width;
    // --- Calculate overall leakage power for unused switch-connections
    leakage_power = 0;


    // (oBs: See FYI calculate_switch_leakage() at the end of this file.) 
    // I can calculate the following two quantities:
    leak_power_in_buffer = switch_inf[1].Leakage_Power;
    leak_power_in_pass_tran = 
      switch_inf[0].Leakage_Power - switch_inf[1].Leakage_Power;
    //printf ("\n leak_power_in_buffer: %g \n", leak_power_in_buffer );
    //printf ("\n leak_power_in_pass_tran: %g \n", leak_power_in_pass_tran );

    //printf ("\n total connections: %d", 
    //	    switch_conn_inside_SB + switch_conn_inside_CB);
    //printf ("\n as sum of switch_conn_inside_SB: %d and switch_conn_inside_CB: %d\n",
    //	    switch_conn_inside_SB, switch_conn_inside_CB);

    //printf ("\n total hwrp connections: %d", total_hwrp_switch_connections);
    //printf ("\n efective_eliminated_due_to_hwrp_insertion: %d", 
    //	    total_efective_eliminated_due_to_hwrp_insertion);
    //printf ("\n USED usual connections: %d", switch_inf[0].num_used);
    //printf ("\n USED hwrp connections: %d", switch_inf[1].num_used);

    // --- Add leak power in USED switch-connections inside SBs and CBs.
    leakage_power +=
      ( switch_inf[0].num_used * leak_power_in_buffer ) +
      // Leak power in "usual" switch-conn buffers. Their pass-tran do
      // not leak. SRAM leak will be added at the end for all switches.
      // Attention:  switch_inf[0].num_used reflects also all used connections
      // inside CBs!
      ( switch_inf[1].num_used * leak_power_in_buffer );
      // If any used hwrp's, then add also leak in their buffers.
      // Attention:  switch_inf[1].num_used can exist only inside SBs.

    //printf ("\n leak power in USED sw-conn: %g \n", leakage_power );

    // --- Add leak power in UN-USED switch-connections inside SBs and CBs.
    leakage_power +=
      ( (switch_conn_inside_SB + switch_conn_inside_CB -
	 total_efective_eliminated_due_to_hwrp_insertion -
	 switch_inf[0].num_used - total_hwrp_switch_connections) 
	* (leak_power_in_buffer + leak_power_in_pass_tran) ) +
      // Leak power in un-used buffers and their pass-tran of (i) usual 
      // switch-connections.
      ( total_hwrp_switch_connections - switch_inf[1].num_used )
      * leak_power_in_buffer;
      // Leak power in un-used buffers of (ii) hwrp switch-connections.
    // printf ("\n + leak power in UN-USED sw-conn: %g \n", leakage_power );

    // --- Add leakage in all SRAMs
    // cristinel.ababei:  Because currently "powermodel" does not include
    // the leakage power of SRAM next summation will be with zero (because
    // SRAM_leakage from .arch file). 
    // This will have to be included to see the more significant
    // impact of inserting hwrp's!
    // In a first approximation we can compute it as the leakage of 
    // two transistors with area 1 and two inverters?
    // based on the schematic shown in V. Betz book in Apendix B. 
    leakage_power +=
      ( switch_conn_inside_SB + switch_conn_inside_CB - 
	total_efective_eliminated_due_to_hwrp_insertion -
	total_hwrp_switch_connections ) * SRAM_leakage;
    //printf ("\n + leak power in SRAMs: %g \n", leakage_power );

    // ----------------------------------------------------------------------

    //Satish. I shall add the leakage power for the SRAM cell also.. Treat the SRAM cells as 6 minimum width transistors (page 208. Appendix B of the book. For hwrps, there will not be an SRAM for the pass transistors inside the switch blocks. Hopefully, this will reduce the power further in case of hwrps

    (*routing_leakage) = leakage_power;
    return(total_routing_power);
}//eof
// ======================================================================= //




/***************************************************************************/
/* Find the probability and transition density for a 2-input MUX */
void find_prob_and_density_for_MUX (
    float probA,
    float probB,
    float probC,
    float densityA,
    float densityB,
    float densityC,
    float *outProb,
    float *outDensity
)
{
    /* Calculations based on the transition density signal model */

    (*outProb) = (1- probA)* probB*probC
                 + probA*(1-probB)*(1-probC)
                 + probA*probB*(1-probC)
                 + probA*probB*probC;

    (*outDensity) = (1 - probC) * densityA + probC * densityB
                    + ((1-probA)*probB +  probA*(1-probB))*densityC;


}

/***************************************************************************/
/* Calculate the Logic Block power dissipation */
static float find_logic_block_power(
    t_subblock_data *subblock_data_ptr,
    float *logic_block_leakage,
    FILE *power_file
)
{
    float total_logic_block_power;
    int lut_size;
    int i, j, k, count;
    int clb_pin_input,  clb_pin_input_num;
    int clb_pin_counter;
    int clock_input, clock_input_index;
    int clock_clb_input[subblock_data_ptr->max_subblocks_per_block];
    float input_and_output_probability[pins_per_clb];
    float input_and_output_density[pins_per_clb];
    float internal_Cwire;
    float previous_LB_power, current_LB_power;
    float unused_CLB_leakage;
    int unused_CLBs;

    /* Assume only one clock input per cluster */
    clock_input = 1;

    total_logic_block_power = 0.0;
    lut_size = subblock_data_ptr->subblock_lut_size;
    internal_Cwire = subblock_data_ptr->clb_Cwire;

    fprintf(power_file, "******** Logic Block Power Analysis  ******* \n");

    /* initialize input and output activities */
    for (clb_pin_counter= 0; clb_pin_counter< pins_per_clb ; clb_pin_counter++)
    {
        input_and_output_probability[clb_pin_counter] = 1.0;
        input_and_output_density[clb_pin_counter] = 0.0;
    }

    /* record input and output activities of the CLB */
    for (i=0; i<num_blocks; i++)
    {
        clock_input_index = -1;
        previous_LB_power = total_logic_block_power;

        if  (subblock_data_ptr->num_subblocks_per_block[i] > 0)
        {
            fprintf(power_file, "\nblock number: %d\n", i);

            /* loop through each look-up-table inside a logic block */
            for(j=0;j<subblock_data_ptr->num_subblocks_per_block[i];j++)
            {
                for (k=0; k<lut_size; k++)
                {
                    clb_pin_input = subblock_data_ptr->subblock_inf[i][j].inputs[k];

                    /* detect a feedback loop in the logic block */
                    if (clb_pin_input > pins_per_clb)
                    {
                        clb_pin_input = clb_pin_input % pins_per_clb;
                    }

                    /* load the probability and the transition density for each input */
                    if (clb_pin_input >=0)
                    {
                        input_and_output_probability[clb_pin_input] =
                            subblock_data_ptr->subblock_inf[i][j].input_probability[k];
                        input_and_output_density[clb_pin_input] =
                            subblock_data_ptr->subblock_inf[i][j].input_density[k];
                    }
                }

                /* load the probability and transition density for the output */
                input_and_output_probability[subblock_data_ptr->subblock_inf[i][j].output] =
                    subblock_data_ptr->subblock_inf[i][j].output_probability;
                input_and_output_density[subblock_data_ptr->subblock_inf[i][j].output] =
                    subblock_data_ptr->subblock_inf[i][j].output_density;

                /* if clock exists, load the probability and density for the clock */
                if (subblock_data_ptr->subblock_inf[i][j].clock != -1)
                {
                    input_and_output_probability[subblock_data_ptr->subblock_inf[i][j].clock] =
                        subblock_data_ptr->subblock_inf[i][j].clock_probability;

                    input_and_output_density[subblock_data_ptr->subblock_inf[i][j].clock] =
                        subblock_data_ptr->subblock_inf[i][j].clock_density;

                    clock_input_index = subblock_data_ptr->subblock_inf[i][j].clock;
                }
            } // end for


            /* Calulate Power due to clock input and the internal wire*/

            if (clock_input_index != -1)
            {
                total_logic_block_power += (0.5*(1+short_circuit_power_percentage)
                                            *swing_voltage*supply_voltage/T_clock_cycle
                                            *input_and_output_density[clock_input_index]
                                            *(2*(INV1x_Cin+INV1x_Cout) + 2*(INV4x_Cin +INV4x_Cout)
                                              +2*internal_Cwire));

            }

            /*
            	Assume the inverters are connected to either ground or VDD when they 
            	are not used and inactive. One of the transistors in each inverter 
            	consumes leakage power.
            */			

            (*logic_block_leakage) += 0.5*(2*(leakage_1X_nmos+leakage_1X_pmos) + 4*leakage_1X_inv);


            /* calculate the power due to the input activity and the 1X and 4X
               inverters and the internal wire */

            clb_pin_input_num = pins_per_clb-clock_input;
            for (clb_pin_counter= 0; clb_pin_counter < clb_pin_input_num; clb_pin_counter++)
            {
                total_logic_block_power += (0.5*(1+short_circuit_power_percentage)
                                            *swing_voltage*supply_voltage/T_clock_cycle
                                            *input_and_output_density[clb_pin_counter]
                                            *(INV1x_Cin + INV1x_Cout + INV4x_Cin + INV4x_Cout
                                              + internal_Cwire));

                /*
                	Assume the inverters are connected to either ground or VDD when they 
                	are not used and inactive. One of the transistors in each inverter 
                	consumes leakage power.
                */
                (*logic_block_leakage) += 0.5* (leakage_1X_inv+leakage_4x_inv);
            }

            for(j=0;j<subblock_data_ptr->num_subblocks_per_block[i];j++)
            {
                /* Power estimation is based on the transition density values
                   internal to the subblock. 
                		
                   each subblock contains:
                	- all the big MUXs (decoders) connected to the input of the LUT
                	- a LUT
                	- a 2 to 1 mux, which select whether the output will go through
                	  the D flip Flop
                	- a Dlip Flop	
                	- 1X and 4X inverters connected to the output  
                */
                total_logic_block_power += power_per_subblock(
                                               clock_input,
                                               lut_size,
                                               subblock_data_ptr->max_subblocks_per_block,
                                               internal_Cwire,
                                               input_and_output_probability,
                                               input_and_output_density,
                                               subblock_data_ptr->subblock_inf[i][j],
                                               logic_block_leakage,
                                               power_file);
            } // end for

            /* if the logic block is not used. then it contributes leakage power */
            unused_CLBs = (subblock_data_ptr->max_subblocks_per_block
                           - subblock_data_ptr->num_subblocks_per_block[i]);

            /*Leakage power per CLBs */
            unused_CLB_leakage = 0.0; /* Added by JL, 2002 */
            calculate_unused_CLB_leakage
            (
                subblock_data_ptr->max_subblocks_per_block,
                lut_size,
                clock_input,
                &unused_CLB_leakage
            );
            unused_CLB_leakage = unused_CLB_leakage * unused_CLBs;

            fprintf (power_file, "Number of unused subblocks: %d\n", unused_CLBs);
            fprintf (power_file, "Leakage power from unused subblocks: %g W\n",
                     unused_CLB_leakage);

            total_logic_block_power += unused_CLB_leakage;
            current_LB_power = total_logic_block_power - previous_LB_power;
            fprintf (power_file, "Power dissipated by this block: %g w\n",
                     current_LB_power);
        } // end if
    } // end for

    fprintf (power_file, "==========================================================\n",
             total_logic_block_power);
    fprintf (power_file, "Total Logic Block energy: %g w\n", total_logic_block_power);

    return(total_logic_block_power);
}

/*****************************************************************/
/* Calculate the power dissipated by each LUT */
float power_per_LUT(
    int lut_size,
    int max_number_of_internal_MUX_per_input,
    float inputProb[],
    float inputDensity[],
    int   inputs[],
    char* function,
    float *outProb,
    float *outDensity,
    float *logic_block_leakage,
    FILE *power_file
)
{
    float internalProb[lut_size][max_number_of_internal_MUX_per_input];
    float internalDensity[lut_size][max_number_of_internal_MUX_per_input];
    float power_per_lut;
    float input_related_power;
    int i,j;
    int inputCounter;
    int MUXs_per_input;
    int MUXCounter;
    int SRAMCounter;

    /* Variables for SRAM function */
    char LUT_SRAM_values[(1<<lut_size)+1];
    int reversed_integer;
    char function_SRAM_values[lut_size +1];
    char reversed_function_SRAM_values[lut_size +1];
    int function_count;


    /* initialize the SRAM_values in the LUTs*/
    /* Notice: the input signals that control the output of
    	the LUT go from LSB to MSB.
    	Please refer to figure B.9 in Vaughn's thesis.
    	In0 is the LSB and IN3 is the MSB. Therefore, the
    	function value needs to be converted to match the
    	LUT SRAM values.  
    	*/

    (*logic_block_leakage) += (1<<lut_size)*SRAM_leakage;

    for (function_count =0;
            function_count < (1<<lut_size);
            function_count++)
    {

        integer_to_SRAMvalues(lut_size,
                              function_count,
                              function_SRAM_values);

        reversed_integer
        = binary_str_to_integer(lut_size,
                                function_SRAM_values);

        LUT_SRAM_values[reversed_integer] = function[function_count];

    }

    LUT_SRAM_values[(1<<lut_size)] = '\0';

    power_per_lut = 0.0;
    input_related_power = 0.0;
    inputCounter = 0;

    for (inputCounter = 0; inputCounter<lut_size;  inputCounter++)
    {

        MUXs_per_input = (1<<(lut_size - 1 - inputCounter));

        input_related_power +=
            0.5*(1+short_circuit_power_percentage)
            *swing_voltage*supply_voltage*inputDensity[inputCounter]/T_clock_cycle
            *(INV1x_Cin
              + (INV1x_Cout + INV2x_Cin)
              + INV2x_Cout
              + INV2x_Cin
              + INV2x_Cout);

        /* one of the transistor in each inverter is leaking during
             the non-transition time */
        (*logic_block_leakage) += 0.5*leakage_1X_inv + leakage_2X_inv;

        MUXCounter = 0;

        if (inputCounter == 0)
        {
            SRAMCounter = 0;
            for (i=0; i<MUXs_per_input; i++)
            {
                find_prob_and_density_for_MUX
                (SRAM_Value_To_float(LUT_SRAM_values[SRAMCounter++]),
                 SRAM_Value_To_float(LUT_SRAM_values[SRAMCounter++]),
                 inputProb[inputCounter],
                 INITIAL_SRAM_DENSITY,
                 INITIAL_SRAM_DENSITY,
                 inputDensity[inputCounter],
                 (&internalProb[inputCounter][i]),
                 (&internalDensity[inputCounter][i]));



                /* power dissipation for the output */

                /* Take care of body effect in the LUT */
                if (swing_voltage >= supply_voltage - NMOS_tx_record.Vth )
                {
                    power_per_lut +=
                        0.5*(1+short_circuit_power_percentage)*(supply_voltage-NMOS_tx_record.Vth)
                        *(supply_voltage-NMOS_tx_record.Vth)/T_clock_cycle
                        * (2*NMOS_C_drain+NMOS_C_gate) * internalDensity[inputCounter][i];
                }
                else
                {
                    /* swing voltage < supply_voltage -NMOS_tx_record.Vth*/
                    power_per_lut +=
                        0.5*(1+short_circuit_power_percentage)*(supply_voltage-NMOS_tx_record.Vth)
                        *swing_voltage/T_clock_cycle
                        * (2*NMOS_C_drain+NMOS_C_gate) * internalDensity[inputCounter][i];
                }

                /* one of the Nmos Pass transistors in the 2-input MUX is not chosen
                	at a given time */
                (*logic_block_leakage) += leakage_1X_nmos;



            } // end for
        } // end if
        else
        {
            for (i=0; i<MUXs_per_input; i++)
            {
                find_prob_and_density_for_MUX
                (internalProb[inputCounter-1][MUXCounter],
                 internalProb[inputCounter-1][MUXCounter+1],
                 inputProb[inputCounter],
                 internalDensity[inputCounter-1][MUXCounter],
                 internalDensity[inputCounter-1][MUXCounter+1],
                 inputDensity[inputCounter],
                 (&internalProb[inputCounter][i]),
                 (&internalDensity[inputCounter][i]));

                /* Take care of body effect in the LUT */
                if (swing_voltage >= supply_voltage - NMOS_tx_record.Vth )
                {
                    /* power dissipation for mux */
                    power_per_lut +=
                        0.5*(1+short_circuit_power_percentage)*(supply_voltage-NMOS_tx_record.Vth)
                        *(supply_voltage-NMOS_tx_record.Vth)/T_clock_cycle
                        * ((NMOS_C_drain * internalDensity[inputCounter-1][MUXCounter])
                           + (NMOS_C_drain * internalDensity[inputCounter-1][MUXCounter+1])
                           + (2*NMOS_C_drain+NMOS_C_gate)* internalDensity[inputCounter][i]);
                }
                else
                {
                    /* swing voltage < supply_voltage -NMOS_tx_record.Vth*/
                    power_per_lut +=
                        0.5*(1+short_circuit_power_percentage)*(supply_voltage-NMOS_tx_record.Vth)
                        *swing_voltage/T_clock_cycle
                        * ((NMOS_C_drain * internalDensity[inputCounter-1][MUXCounter])
                           + (NMOS_C_drain * internalDensity[inputCounter-1][MUXCounter+1])
                           + (2*NMOS_C_drain+NMOS_C_gate)* internalDensity[inputCounter][i]);
                }

                /* one of the Nmos Pass transistors in the 2-input MUX is not chosen
                	at a given time */
                (*logic_block_leakage) += leakage_1X_nmos;

                MUXCounter +=2;
            }
        }

    } // end for

    (*outProb) =  internalProb[lut_size -1][0];
    (*outDensity) = internalDensity[lut_size -1][0];

    return(power_per_lut + input_related_power);

}


/*****************************************************************/
/* convert the SRAM_value to floating point*/
float SRAM_Value_To_float(
    char SRAM_value
)
{
    if (SRAM_value == '1')
    {
        return(1.0);
    }
    else
    {
        return(0.0);
    }
}

/*****************************************************************/
/* Calculate the capacitance for componesnts that are commonly
   used within the logic block;
	- minimum size NMOS and PMOS transistors
	- 1X, 2X, and 4X inverters 
 */

void calculate_individual_cap (void)
{
    /* minimum size NMOS */
    find_cap(NMOS,
             NMOS_tx_record.min_length,
             NMOS_tx_record.min_width,
             &NMOS_C_drain,
             &NMOS_C_source,
             &NMOS_C_gate);

    /* minimum size PMOS */
    find_cap(PMOS,
             PMOS_tx_record.min_length,
             PMOS_tx_record.min_width,
             &PMOS_C_drain,
             &PMOS_C_source,
             &PMOS_C_gate);


    /* 1X inverter */
    calculate_INV_cap(PMOS_tx_record.min_length,
                      PMOS_tx_record.min_width*1.9,
                      NMOS_tx_record.min_length,
                      NMOS_tx_record.min_width,
                      &INV1x_Cin,
                      &INV1x_Cout);

    /* 2X inverter */
    calculate_INV_cap(PMOS_tx_record.min_length,
                      (PMOS_tx_record.min_width*2*1.9),
                      NMOS_tx_record.min_length,
                      (NMOS_tx_record.min_width*2),
                      &INV2x_Cin,
                      &INV2x_Cout);

    /* 4X inverter */
    calculate_INV_cap
    (PMOS_tx_record.min_length,
     (PMOS_tx_record.min_width*1.9*4),
     NMOS_tx_record.min_length,
     (NMOS_tx_record.min_width*4),
     &INV4x_Cin,
     &INV4x_Cout);


}

/*****************************************************************/
/* Calculate the gate capacitance */
float calculate_gate_cap(
    float COX,
    float length,
    float width
)
{
    float gate_capacitance;
    float poly_length;


    /* Gate capacitance is equal to oxide capacitance plus the
         poly capacitance */
    gate_capacitance = 0.0;

    if (width <= size_limit*NMOS_tx_record.min_width )
    {
        poly_length
        = 2*poly_inf.poly_extention + width;
    }
    else
    {
        poly_length
        = 3*length + 2*(width/2+poly_inf.poly_extention+length);
    }

    gate_capacitance
    = poly_length*poly_inf.Cpoly+COX*width*length;

    return (gate_capacitance);
}


/*****************************************************************/
/* Calculate the drain_source capacitance for transistor that
   is smaller than the size limit */
void find_drain_source_cap_small(
    float length,
    float width,
    float* C_drain,
    float* C_source,
    float CJ,
    float CJSW,
    float CJSWG,
    float CGDO
)
{
    (*C_drain) = 2.5*length*width*CJ + (5*length+width)*CJSW
                 + width*(CGDO+CJSWG);

    (*C_source) = 2.5*length*width*CJ + (5*length+width)*CJSW
                  + width*(CGDO+CJSWG);
}


/*****************************************************************/
/* Calculate the drain_source capacitance for transistor that
   is larger than the size limit */
void find_drain_source_cap_large(
    float length,
    float width,
    float* C_drain,
    float* C_source,
    float CJ,
    float CJSW,
    float CJSWG,
    float CGDO
)
{

    (*C_drain) = 2.5*length*(width/2)*CJ + 5*length*CJSW + width*(CGDO+CJSWG);

    (*C_source) = 2*(2.5*length*(width/2)*CJ + (5*length+(width/2))*CJSW
                     + (width/2)*(CGDO+CJSWG));

}

/*****************************************************************/
/* Calculate the capacitance for a transistor */
void find_cap(
    tx_type transistor_type,
    float length,
    float width,
    float * C_drain,
    float * C_source,
    float * C_gate
)
{
    float CJ;
    float CJSW;
    float CJSWG;
    float CGDO;
    float COX;

    if ( transistor_type == NMOS)
    {
        /* load NMOS values */
        CJ = NMOS_tx_record.CJ;
        CJSW = NMOS_tx_record.CJSW;
        CJSWG = NMOS_tx_record.CJSWG;
        CGDO =NMOS_tx_record.CGDO;
        COX = NMOS_tx_record.COX ;
    }
    else
    {
        /* load PMOS values */
        CJ = PMOS_tx_record.CJ;
        CJSW = PMOS_tx_record.CJSW;
        CJSWG = PMOS_tx_record.CJSWG;
        CGDO =PMOS_tx_record.CGDO;
        COX = PMOS_tx_record.COX ;
    }

    /* Capacitance Calculation */
    (*C_gate)  = calculate_gate_cap(COX, length, width);

    /* check if the transistor is larger than the size limit */
    if (width >= size_limit*NMOS_tx_record.min_width)
    {
        find_drain_source_cap_large(
            length,
            width,
            C_drain,
            C_source,
            CJ,
            CJSW,
            CJSWG,
            CGDO
        );
    }
    else
    {
        find_drain_source_cap_small(
            length,
            width,
            C_drain,
            C_source,
            CJ,
            CJSW,
            CJSWG,
            CGDO
        );
    }

}

/*****************************************************************/
/* Calculate the input and output capacitance of an inverter */
void calculate_INV_cap(
    float PMOS_length,
    float PMOS_width,
    float NMOS_length,
    float NMOS_width,
    float *Cin,
    float *Cout
)
{
    float inv_NMOS_Cgate, inv_NMOS_Cdrain, inv_NMOS_Csource;
    float inv_PMOS_Cgate, inv_PMOS_Cdrain, inv_PMOS_Csource;

    find_cap(NMOS,
             NMOS_length,
             NMOS_width,
             &inv_NMOS_Cdrain,
             &inv_NMOS_Csource,
             &inv_NMOS_Cgate);

    find_cap(PMOS,
             PMOS_length,
             PMOS_width,
             &inv_PMOS_Cdrain,
             &inv_PMOS_Csource,
             &inv_PMOS_Cgate);

    (*Cin) =  inv_NMOS_Cgate + inv_PMOS_Cgate;
    (*Cout) =  inv_NMOS_Cdrain + inv_PMOS_Cdrain;

}



/***************************************************************************/
/* Calculate the power of a DFF */
float calculate_DFF_power(
    float inputProb,
    float inputDensity,
    float clockProb,
    float clockDensity,
    float* outProb,
    float* outDensity,
    FILE *power_file
)
{
    float DFF_power;
    float effective_density;
    int increase_factor;

    DFF_power = 0.0;

    /* Equation for DFF is derived by Matlab */
    /* Power = (-0.074)*Density + (5.2486)*Density*Density */

    if (inputDensity <= (clockDensity/2))
    {
        effective_density
        = (-0.074)*inputDensity + (5.2486)*inputDensity*inputDensity;
    }
    else
    {
        /* Special case:
         when input density is higher than half of the clock frequency
         the DFF will try to settle down the glitches. Model as
               an additional linear factor relative to half of the clock
         density.
        */	 
        increase_factor
        = 1+ (inputDensity - clockDensity/2)/2;

        effective_density
        =((-0.074)*(clockDensity/2)+ (5.2486)*(clockDensity/2)*(clockDensity/2))*increase_factor;
    }


    /* Power Calculation */

    DFF_power
    += (0.5*swing_voltage*supply_voltage/T_clock_cycle*effective_density
        * (8*(NMOS_C_drain +   PMOS_C_drain)
           + 4*(1+short_circuit_power_percentage)*(INV1x_Cin+INV1x_Cout)
           + 4*(NMOS_C_gate+PMOS_C_gate)));

    (*outProb) = inputProb;
    (*outDensity) = effective_density;

    return (DFF_power);
}

/***************************************************************************/
/* Calculate power dissipated inside the subblock */
float power_per_subblock(
    int clock_input,
    int lut_size,
    int max_subblock_per_block,
    float internal_Cwire,
    float input_and_output_probability[],
    float input_and_output_density[],
    t_subblock subblock_ptr,
    float *logic_block_leakage,
    FILE *power_file
)
{
    int level;
    int check_level;
    int i,j, k, temp;
    char SRAM_values[20];
    int LUT_input, LUT_input_counter;
    float temp_LUT_input_prob[lut_size], LUT_input_prob[lut_size];
    float temp_LUT_input_density[lut_size], LUT_input_density[lut_size];
    float LUT_output_prob, LUT_output_density;
    float subblock_clock_prob;
    float subblock_clock_density;
    float D_FF_Cin, D_FF_Cout;
    float D_FF_output_prob, D_FF_output_density;
    float energy_of_DFF;
    float power_of_this_subblock;
    int max_number_of_internal_MUX_per_input;
    char SRAM_2to1MUX;
    float output_prob_for_2to1MUX, output_density_for_2to1MUX;

    /* Power estimation is based on the transition density values
    	internal to the subblock. 

    	each subblock contains:
    	 - all the big MUXs (decoders) connected to the input of the LUT
    	 - a LUT
    	 - a 2 to 1 mux, which select whether the output will go through
    	   a D flip Flop
    	 - a Dlip Flop	
    	 - 1X and 4X inverters connected to the output  
     */

    fprintf (power_file, "subblock name:%s \t\t subblock power: ",  subblock_ptr.name);

    power_of_this_subblock = 0.0;

    /* Calculate how big the MUXs (decoders) need to be */
    /* the level represents the number of levels of 2-input MUX
    required for the big MUXs*/
    check_level = pins_per_clb - clock_input;
    level = 0;

    while (check_level > 0)
    {
        check_level = check_level>>1;
        level ++;
    }

    /* Calculate power dissipation due to the input MUXs
    	and the 1X inverter at the end of the MUX. */

    LUT_input = 0;
    for (i=0; i<lut_size; i++)
    {
        if (subblock_ptr.inputs[i] != -1)
        {
            find_SRAMvalues_for_input_MUX
            (level,
             subblock_ptr.inputs[i],
             max_subblock_per_block,
             clock_input,
             SRAM_values);

            LUT_input++;
        }
        else
        {
            /* Assume the unused input connected to the first input to the MUX */
            find_SRAMvalues_for_input_MUX
            (level,
             0,
             max_subblock_per_block,
             clock_input,
             SRAM_values);
        }

        (*logic_block_leakage) += level * SRAM_leakage;

        /* Calculate the power dissipation for the big MUX */
        power_of_this_subblock
        += power_per_input_MUX
           (pins_per_clb - clock_input,
            pins_per_clb - max_subblock_per_block,
            input_and_output_probability,
            input_and_output_density,
            level,
            SRAM_values,
            logic_block_leakage,
            &temp_LUT_input_prob[i],
            &temp_LUT_input_density[i]);



        /* Calculate the power dissipated by the 1X inverter after the big MUX */
        power_of_this_subblock
        += (0.5*(1+short_circuit_power_percentage)
            *swing_voltage*supply_voltage*temp_LUT_input_density[i]/T_clock_cycle
            *(INV1x_Cin + INV1x_Cout));


        /* one of the transistors in the inverter is leaking during
         the non-transition time */
        (*logic_block_leakage) += 0.5*leakage_1X_inv;

    } // end for

    /* Get the input probability for LUT */
    /* Notice the inputs to LUT have been re-ordered since the
     first input to the LUT is the least significant bit and
     the last input is the MSB.
     */

    for (k=0; k<LUT_input; k++)
    {
        LUT_input_prob[k] = temp_LUT_input_prob[LUT_input - k - 1];
        LUT_input_density[k] = temp_LUT_input_density[LUT_input - k -1];
    }

    LUT_input_counter = LUT_input;

    while (LUT_input_counter < lut_size)
    {
        /* set the unused input probability to 1.0*/
        LUT_input_prob[LUT_input_counter] = 1.0;
        LUT_input_density[LUT_input_counter++] = 0.0;
    }

    /* Calculate Power dissipation due to the LUT */

    max_number_of_internal_MUX_per_input = (1<<lut_size) >> 1;

    power_of_this_subblock
    += power_per_LUT
       (lut_size,
        max_number_of_internal_MUX_per_input,
        LUT_input_prob,
        LUT_input_density,
        subblock_ptr.inputs,
        subblock_ptr.function,
        &LUT_output_prob,
        &LUT_output_density,
        logic_block_leakage,
        power_file);

    /* replace the activity values from the input files */
    LUT_output_prob = subblock_ptr.between_lut_and_ff_probability;
    LUT_output_density = subblock_ptr.between_lut_and_ff_density;

    /* Calculate the power dissipation due to D-ff */
    /* if the D-FF is used, calculate the power within the D-ff;
     otherwise, the input capacitance of D-ff still needs to 
     be included in the calculation. */

    if (subblock_ptr.clock == -1)
    {
        subblock_clock_prob = 0.0;
        subblock_clock_density = 0.0;
        D_FF_output_prob = 0.0;
        D_FF_output_density =0.0;

        (*logic_block_leakage)
        += (2*leakage_1X_nmos
            + 2*leakage_1X_pmos
            + 4*leakage_1X_inv);
    }
    else
    {
        subblock_clock_prob = subblock_ptr.clock_probability;
        subblock_clock_density = subblock_ptr.clock_density;

        power_of_this_subblock
        += calculate_DFF_power(LUT_output_prob,
                               LUT_output_density,
                               subblock_clock_prob,
                               subblock_clock_density,
                               &D_FF_output_prob,
                               &D_FF_output_density,
                               power_file
                              );

    }


    /* Calculate the power due to 2:1 MUX at the end of the LUT
     output*/
    /* Assume the 2:1 MUX use min-size transistor */

    if (subblock_ptr.clock == -1)
    {
        SRAM_2to1MUX = '0';
    }
    else
    {
        SRAM_2to1MUX = '1';
    }

    find_prob_and_density_for_MUX
    (LUT_output_prob,
     D_FF_output_prob,
     SRAM_Value_To_float(SRAM_2to1MUX),
     LUT_output_density,
     D_FF_output_density,
     INITIAL_SRAM_DENSITY,
     &output_prob_for_2to1MUX,
     &output_density_for_2to1MUX);

    /* power dissipated at the inputs of the 2to1 MUX*/
    power_of_this_subblock
    += 0.5*(1+short_circuit_power_percentage)
       *swing_voltage*supply_voltage/T_clock_cycle
       *D_FF_output_density*NMOS_C_drain;

    power_of_this_subblock
    += (0.5*(1+short_circuit_power_percentage)
        *swing_voltage*supply_voltage/T_clock_cycle
        *LUT_output_density*NMOS_C_drain);

    /* replace output_density_for_2to1MUX with the value from input file */
    output_density_for_2to1MUX = subblock_ptr.output_density;

    /* power dissipated at the output of the 2to1 MUX */
    power_of_this_subblock
    += (0.5*(1+short_circuit_power_percentage)
        *swing_voltage*supply_voltage/T_clock_cycle
        *output_density_for_2to1MUX*(2*NMOS_C_drain+NMOS_C_gate));

    /* calculate leakage power of the 2 to 1 mux*/
    (*logic_block_leakage) += SRAM_leakage;

    /* one transistor in the 2-input MUXis not used */
    (*logic_block_leakage) += leakage_1X_nmos;


    /* Power dissipation due to the 1X and 4X inverters and the internal
     wire at the end */
    power_of_this_subblock
    += (0.5*(1+short_circuit_power_percentage)*swing_voltage*supply_voltage
        *output_density_for_2to1MUX/T_clock_cycle
        *(INV1x_Cin+INV1x_Cout+INV4x_Cin+INV4x_Cout)
        + internal_Cwire);

    fprintf (power_file,
             "%g W\n", power_of_this_subblock);
    return( power_of_this_subblock);
}

/***************************************************************************/
/*calculate the power dissipation for a MUX*/
float power_per_input_MUX(
    int overall_input_to_MUX,
    int inputnum,
    float inputProb[],
    float inputDensity[],
    int level,
    char  SRAM_values[],
    float *logic_block_leakage,
    float *outProb,
    float *outDensity
)
{
    float internal_prob[level][overall_input_to_MUX];
    float internal_density[level][overall_input_to_MUX];
    int current_level, i, j, k;
    int MUX_num_per_level, previous_MUX_num_per_level, MUX_counter;
    int level_counter;
    float total_input_MUX_energy;
    float SRAM_numerical_value;

    total_input_MUX_energy = 0.0;
    MUX_num_per_level = overall_input_to_MUX;


    /* Power due to the input capacitance of the MUX */
    for (i=0; i< inputnum; i++)
    {
        total_input_MUX_energy
        += 0.5*(1+short_circuit_power_percentage)
           *swing_voltage*supply_voltage/T_clock_cycle
           *inputDensity[i]*(NMOS_C_drain);
    }

    /* loop through each level */
    for (current_level=0; current_level<level; current_level++)
    {
        previous_MUX_num_per_level = MUX_num_per_level;

        if (MUX_num_per_level%2 == 0)
        {
            MUX_num_per_level = MUX_num_per_level/2;
            MUX_counter = MUX_num_per_level;
        }
        else
        {
            MUX_num_per_level = MUX_num_per_level/2+1;
            MUX_counter = MUX_num_per_level-1;
            if (current_level == 0)
            {
                internal_prob[current_level][MUX_num_per_level-1]
                = inputProb[previous_MUX_num_per_level-1];
                internal_density[current_level][MUX_num_per_level-1]
                = inputDensity[previous_MUX_num_per_level-1];
            }
            else
            {
                internal_prob[current_level][MUX_num_per_level-1]
                = internal_prob[current_level-1][previous_MUX_num_per_level-1];
                internal_density[current_level][MUX_num_per_level-1]
                =internal_density[current_level-1][previous_MUX_num_per_level-1];

            }
        }

        level_counter =0;

        SRAM_numerical_value = SRAM_Value_To_float(SRAM_values[current_level]);

        if (current_level == 0)
        {
            for (k=0; k<MUX_counter;k++)
            {

                /* calculate the probability and density at the output of
                	each MUX */
                find_prob_and_density_for_MUX
                (inputProb[level_counter],
                 inputProb[level_counter+1],
                 SRAM_numerical_value,
                 inputDensity[level_counter],
                 inputDensity[level_counter+1],
                 INITIAL_SRAM_DENSITY,
                 &internal_prob[current_level][k],
                 &internal_density[current_level][k]);

                total_input_MUX_energy
                += 0.5*(1+short_circuit_power_percentage)
                   *swing_voltage*supply_voltage
                   *(2*NMOS_C_drain + NMOS_C_gate)
                   *internal_density[current_level][k]/T_clock_cycle;

                /* One of the NMOS transistor in the 2-input MUX is not chosen at
                   any given time*/					
                (*logic_block_leakage) += leakage_1X_nmos;

                level_counter += 2;
            }
        }
        else
        {
            for (k=0; k<MUX_counter;k++)
            {
                find_prob_and_density_for_MUX
                (internal_prob[current_level-1][level_counter],
                 internal_prob[current_level-1][level_counter+1],
                 SRAM_numerical_value,
                 internal_density[current_level-1][level_counter],
                 internal_density[current_level-1][level_counter+1],
                 INITIAL_SRAM_DENSITY,
                 &internal_prob[current_level][k],
                 &internal_density[current_level][k]);

                /* power dissipation of the MUX input */
                total_input_MUX_energy
                += (0.5*(1+short_circuit_power_percentage)
                    *swing_voltage*supply_voltage/T_clock_cycle
                    *((NMOS_C_drain*internal_density[current_level-1][level_counter])
                      + (NMOS_C_drain*internal_density[current_level-1][level_counter+1])));


                /* power dissipation of the MUX output */
                total_input_MUX_energy
                += (0.5*(1+short_circuit_power_percentage)
                    *swing_voltage*supply_voltage/T_clock_cycle
                    *(2*NMOS_C_drain + NMOS_C_gate)*internal_density[current_level][k]);

                /* One of the NMOS transistors in the 2-input MUX is not chosen at
                   any given time*/	
                (*logic_block_leakage) += leakage_1X_nmos;

                level_counter += 2;
            }
        }

    }

    (*outProb) = internal_prob[level-1][0];
    (*outDensity) = internal_density[level-1][0];
    total_input_MUX_energy *= correlation_factor;

    return(total_input_MUX_energy);
}

/***************************************************************************/
/*  Change the selected input to a stream of binary value     */
void integer_to_SRAMvalues(
    int SRAM_num,
    int input_integer,
    char SRAM_values[]
)
{
    char binary_str[20];
    int binary_str_counter;
    int local_integer;
    int i;

    binary_str_counter = 0;

    local_integer = input_integer;

    while (local_integer >0)
    {
        if (local_integer %2 == 0 )
        {
            SRAM_values[binary_str_counter++]  = '0';
        }
        else
        {
            SRAM_values[binary_str_counter++]  = '1';
        }
        local_integer = local_integer /2;
    }

    while (binary_str_counter < SRAM_num)
    {
        SRAM_values[binary_str_counter++]  = '0';
    }

    SRAM_values[binary_str_counter]  = '\0';

    for (i=0; i<binary_str_counter; i++)
    {
        binary_str[i] = SRAM_values[binary_str_counter-1-i];
    }

    binary_str[binary_str_counter]  = '\0';


}

/***************************************************************************/
void find_SRAMvalues_for_input_MUX(
    int SRAM_num,
    int input_clb_pin_number,
    int max_subblock_per_block,
    int clock_input,
    char SRAM_values[]
)
{
    int output_select;

    if (input_clb_pin_number > pins_per_clb)
    {
        /* this is a feedback loop */
        /* Note:
             	pins_per_clb 
             = number of input pin + number of outputs + clock input 

         If the output of a subblock connects to an input of the same 
         subblock, its pin number is pins_per_clb + the output pin index.
        */
        output_select
        = input_clb_pin_number - max_subblock_per_block - clock_input;
    }
    else
    {
        /* this is a normal input */
        output_select = input_clb_pin_number;
    }

    /* convert the input to binary */
    integer_to_SRAMvalues(SRAM_num,
                          output_select,
                          SRAM_values);

}

/***************************************************************************/
/* convert a binary bit stream back to an integer */
int binary_str_to_integer(
    int binary_length,
    char* binary_str
)
{
    int i;
    int integer_value;
    int temp_value;

    integer_value = 0;

    for (i=0; i<binary_length; i++)
    {
        if (binary_str[i] == '1')
        {
            temp_value = (1<<(binary_length-i-1));
            integer_value += temp_value;
        }
    }

    return (integer_value);
}
/***************************************************************************/
/* Reverse a binary string */
void reverse_binary_str(
    int binary_length,
    char* binary_str,
    char* reverse_binary_str
)
{
    int i;

    for (i=0; i<binary_length; i++)
    {
        reverse_binary_str[i] = binary_str[binary_length-1-i];
    }
    reverse_binary_str[binary_length] = '\0';
}
/***************************************************************************/
/* Check if the net is a global clock network */
static boolean is_global_clock(
    int inet,
    int *num_subblocks_per_block,
    t_subblock **subblock_inf,
    float *clock_density
)
{

    /* Returns TRUE if the net driven by this block (which must be an INPAD) is  *
     * (1) a global signal, and (2) used as a clock input to at least one block. */

    int  ipin, to_blk, to_pin, isub;
    (*clock_density) = 0;

    for (ipin=1;ipin<net[inet].num_pins;ipin++) {
        to_blk = net[inet].blocks[ipin];
        to_pin = net[inet].blk_pin[ipin];

        for (isub=0;isub<num_subblocks_per_block[to_blk];isub++) {
            if (subblock_inf[to_blk][isub].clock == to_pin)
            {
                (*clock_density) = subblock_inf[to_blk][isub].clock_density;
                return (TRUE);
            }
        }
    }

    return (FALSE);
}

/***************************************************************************/
/* calculate the power dissipation for the clock */
float find_clock_power(
    float *clock_leakage,
    FILE *power_file
)
{
    float Cload_for_clock_per_clb;
    float Clock_power_dissipation;
    int total_clock_buffers, total_clock_segments;
    int i;

    /* Initialization */
    Clock_power_dissipation = 0;
    total_clock_buffers = 0;
    total_clock_segments =0;

    fprintf (power_file, "************ Clock Power Analysis ************\n\n");

    /* if no global clock, then return */
    if (num_global_clock == 0)
    {
        return(0.0);
    }

    for (i=0; i<num_global_clock; i++)
    {

        /* Assume the global clock is active even for combinational circuits */
        if (num_global_clock == 1)
        {
            if (clock_inf[i].clock_density == 0)
            {
                clock_inf[i].clock_density = 2;
            }
        }
        /* find the power dissipated by each clock network */
        Clock_power_dissipation
        +=  find_clock_power_per_network (clock_leakage,
                                          i,
                                          &total_clock_buffers,
                                          &total_clock_segments,
                                          power_file);

    }

    return (Clock_power_dissipation);
}


/***************************************************************************/
/* Calculate the power dissipation per clock network */
float find_clock_power_per_network(
    float *clock_leakage,
    int clock_index,
    int* total_clock_buffers,
    int* total_clock_segments,
    FILE *power_file
)
{
    float clock_network_energy, clock_network_delay;
    int i, level_needed;
    float buffer_needed_per_segment_length;
    float power_per_level, delay_per_level;
    float buffer_delay, segment_delay;
    int avg_num_of_branches;
    int buffers_with_extra_branches, extra_branches, leftover_clbs;
    float leaf_wire_cap, extra_segment_length;
    int buffers_with_no_branches;
    int num_leaf_buffers;
    int clock_buffer_size;
    int incremental_segment_length;

    power_per_level =0;
    delay_per_level =0;
    incremental_segment_length =0;

    /* Estimate the relative size of the clock buffer */
    clock_buffer_size
    = (clock_inf[clock_index].buffer_Cout)/INV1x_Cout;

    /* Calculate the leakage power for the clock buffer */
    leakage_clock_buffer
    = calculate_INV_leakage(clock_buffer_size,
                            clock_inf[clock_index].buffer_Cout);

    /* Assume the FPGA is in square shape */
    /* the number of branches will double per layer */
    level_needed = log(nx*ny)/log(2);

    /* Calculate the RC delay of each clock buffer */
    buffer_delay
    = clock_inf[clock_index].buffer_R
      *(clock_inf[clock_index].buffer_Cin
        + clock_inf[clock_index].buffer_Cout);

    /* Calculate the delay of this wire segement */
    segment_delay
    = clock_inf[clock_index].Rwire*clock_inf[clock_index].Cwire/2;

    /* Calculate the additional buffers required per unit segment */
    buffer_needed_per_segment_length = sqrt(segment_delay/buffer_delay);

    /* Initialization */
    clock_network_energy = 0.0;
    clock_network_delay = 0.0;

    for (i=level_needed-1; i>=0;i--)
    {
        if (i == level_needed-1)
        {
            /* for the first 3 bottom levels,
                length of the base segment always spans half 
             of a logic block */
            incremental_segment_length = 0.5;

        }
        else if (((level_needed-1)-i-1)%2 == 0)
        {
            /* segment length increases every two levels
             starting from the second bottom level*/
            incremental_segment_length *= 2;
        }

        /* calculate the power dissipation per level */
        find_clock_power_per_level
        (
            clock_index,
            i,
            incremental_segment_length,
            buffer_delay,
            segment_delay,
            buffer_needed_per_segment_length,
            &delay_per_level,
            &power_per_level,
            clock_leakage,
            total_clock_buffers,
            total_clock_segments
        );

        clock_network_energy += power_per_level;
        clock_network_delay += delay_per_level;

    }
    /* At the top level, there is one long segment from the PLL to the
    	center of the FPGA.  */
    incremental_segment_length = nx/2;
    find_clock_power_per_level
    (
        clock_index,
        0,
        incremental_segment_length,
        buffer_delay,
        segment_delay,
        buffer_needed_per_segment_length,
        &delay_per_level,
        &power_per_level,
        clock_leakage,
        total_clock_buffers,
        total_clock_segments
    );
    clock_network_energy += power_per_level;
    clock_network_delay += delay_per_level;

    /* Consider the very bottom layer of the clock network */

    num_leaf_buffers =  (1<<(level_needed-1));

    /* round up the average number of branches at each leaf */
    avg_num_of_branches = (nx*ny) / num_leaf_buffers + 0.5;

    /* Assume each wire at the leaf spans half of a logic block */
    leaf_wire_cap
    = num_leaf_buffers*(clock_inf[clock_index].Cwire
                        *avg_num_of_branches*0.5);

    clock_network_energy
    += 0.5*swing_voltage*supply_voltage*(1+short_circuit_power_percentage)
       /T_clock_cycle*leaf_wire_cap*clock_inf[clock_index].clock_density;


    clock_network_delay
    +=  pow(0.5,2)*segment_delay;


    (*total_clock_segments)
    += num_leaf_buffers*(avg_num_of_branches*0.5);

    fprintf (power_file,"Clock Network %d: Power= %g, Delay=%g\n",
             clock_index, clock_network_energy, clock_network_delay);

    return (clock_network_energy);
}

/***************************************************************************/
/* Calculate the power dissipation at each level of the clock network */
void find_clock_power_per_level(
    int clock_index,
    int current_level,
    int length_of_segment_at_this_level,
    float buffer_delay,
    float segment_delay,
    float buffer_needed_per_segment_length,
    float *total_delay_at_this_level,
    float *total_power_at_this_level,
    float *clock_leakage,
    int *total_clock_buffers,
    int *total_clock_segments
)
{
    int multiple_clock_segments;
    int number_of_buffers_inserted, number_of_chopped_segment ;
    float length_of_chopped_clock_segment, overall_segment_length;
    float total_capacitance;
    int seg_vector;

    /* Initialize */
    (*total_delay_at_this_level) =0;
    (*total_power_at_this_level)=0;

    /* Calculate number of clock segments in this level */
    /* assuming the clock segments at the same level have
    	the same length */
    multiple_clock_segments = (1<<current_level);

    number_of_buffers_inserted
    = length_of_segment_at_this_level*buffer_needed_per_segment_length;

    number_of_chopped_segment = number_of_buffers_inserted +1;

    length_of_chopped_clock_segment
    = length_of_segment_at_this_level/number_of_chopped_segment;


    /* Delay for this segment is roughly equal to the total of
    	buffer delay and the segment delay */
    (*total_delay_at_this_level)
    += (number_of_buffers_inserted+1)*buffer_delay
       + number_of_chopped_segment
       * pow(length_of_chopped_clock_segment,2) * segment_delay;


    total_capacitance
    = multiple_clock_segments
      * ((number_of_buffers_inserted+1)*(clock_inf[clock_index].buffer_Cin
                                         + clock_inf[clock_index].buffer_Cout)
         + number_of_chopped_segment*(length_of_chopped_clock_segment
                                      * clock_inf[clock_index].Cwire));


    (*total_power_at_this_level)
    = (0.5*swing_voltage*supply_voltage*(1+short_circuit_power_percentage)
       /T_clock_cycle
       *total_capacitance
       *clock_inf[clock_index].clock_density);

    /* if clock is not used, calculate the leakage power */
    /* Can be used in multiple global clock network, which is not
         supported now. */
    if (clock_inf[clock_index].clock_density == 0)
    {
        (*clock_leakage)
        += (multiple_clock_segments
            *(number_of_buffers_inserted+1)
            *leakage_clock_buffer);
    }

    (*total_clock_buffers)
    += multiple_clock_segments*(number_of_buffers_inserted+1);

    (*total_clock_segments)
    += multiple_clock_segments*length_of_segment_at_this_level;
}

/***************************************************************************/
/* Calculate the leakage power for unused CLBs */
void calculate_unused_CLB_leakage (
    int maximum_subblocks_per_block,
    int lut_size,
    int clock_input,
    float *logic_block_leakage
)
{
    int num_inputs_per_block;
    int SRAM_needed_for_MUX;

    num_inputs_per_block = pins_per_clb - clock_input;

    /* 1X and 4X inverters linked to the inputs */
    (*logic_block_leakage)
    += num_inputs_per_block*(leakage_1X_inv+leakage_4x_inv);

    /* 1X and 4X inverters linked to the outputs */
    (*logic_block_leakage)
    += maximum_subblocks_per_block*(leakage_1X_inv+leakage_4x_inv);

    /* DFF leakage */
    (*logic_block_leakage)
    += maximum_subblocks_per_block*(4*leakage_1X_nmos+4*leakage_1X_pmos + 2*leakage_1X_inv);

    /* LUT leakage */
    (*logic_block_leakage)
    += maximum_subblocks_per_block
       *((1<<lut_size)*SRAM_leakage
         + 2*((1<<lut_size)-1)*leakage_1X_nmos
         + lut_size*(leakage_1X_inv + 2*leakage_2X_inv));

    /* 2 to 1 mux at the output of the lut */
    (*logic_block_leakage)
    += maximum_subblocks_per_block*(2*leakage_1X_nmos + SRAM_leakage);

    /* MUX leakage */
    SRAM_needed_for_MUX
    = log(num_inputs_per_block+maximum_subblocks_per_block)/log(2);

    (*logic_block_leakage)
    +=  maximum_subblocks_per_block
        * (SRAM_needed_for_MUX*SRAM_leakage
           + 2*((1<<SRAM_needed_for_MUX)-1)*leakage_1X_nmos);

}

/***************************************************************************/
/* Calculate the subthreshold current for a transistor */
float calculate_leakage_power (
    tx_type transistor_type,
    float diffusion_cap,
    int size
)
{
    int temperature_in_K;
    float NFS, COX, Nleak, VTH, EC;
    float nKT_overQ, Von;
    float Ion, Ioff;
    float min_width, Leff;
    float leakage_power;

    if (transistor_type == NMOS)
    {
        NFS = get_NFS(NMOS, operation_temperature);
        COX = NMOS_tx_record.COX;
        VTH = NMOS_tx_record.Vth;
        EC = NMOS_tx_record.EC;
        Leff = NMOS_tx_record.min_length;
        min_width = NMOS_tx_record.min_width;
    }
    else
    {
        NFS = get_NFS(PMOS, operation_temperature);
        COX = PMOS_tx_record.COX;
        VTH = PMOS_tx_record.Vth;
        EC = PMOS_tx_record.EC;
        Leff = PMOS_tx_record.min_length;
        min_width = PMOS_tx_record.min_width;
    }


    temperature_in_K = operation_temperature + 273;
    Nleak = 1 + (Q_charge*NFS/COX) + (diffusion_cap/COX);

    nKT_overQ = Nleak*Bolzmann_constant*temperature_in_K/Q_charge;
    Von = VTH + nKT_overQ;
    Ion = size* min_width*COX*vsat*(pow(nKT_overQ,2))/(nKT_overQ+EC*Leff);
    Ioff = Ion*exp((Vgs_for_leakage-Von)/nKT_overQ);

    /* leakage power = subthreshold current * supply_voltage */
    leakage_power = Ioff*supply_voltage;
    return(leakage_power);
}


/***************************************************************************/
/* Get the NFS value for the operation temperature */
float get_NFS(
    tx_type transistor_type,
    int temperature
)
{
    int i;
    int min_diff;
    int diff_index;

    for (i=0; i<num_temperature_records; i++)
    {
        /* if found, output the NFS value directly */
        if (NFS_records[i].temperature == temperature)
        {
            if (transistor_type == NMOS)
                return (NFS_records[i].NMOS_NFS_value);
            else
                return (NFS_records[i].PMOS_NFS_value);
        }

        /* if not yet found, check the difference between
        	the temperature and the one in the record.*/
        if (i ==0)
        {
            min_diff = abs(NFS_records[i].temperature - temperature);
            diff_index = i;
        }
        else
        {
            if (min_diff > abs(NFS_records[i].temperature - temperature))
            {
                min_diff = abs(NFS_records[i].temperature - temperature);
                diff_index = i;
            }
        }
    }

    /* if the operation temperature is not found,
    then use the NFS of the closest temperature 
    available in the record.
    */
    if (transistor_type == NMOS)
        return (NFS_records[diff_index].NMOS_NFS_value);
    else
        return (NFS_records[diff_index].PMOS_NFS_value);

}

/***************************************************************************/
/* Calculate the leakage power for a inverter */
float calculate_INV_leakage(
    int inverter_size,
    float diffusion_cap
)
{
    float leakage_power;
    float PMOS_diffusion;
    float NMOS_diffusion;

    leakage_power = 0.0;

    PMOS_diffusion = (P_to_N_size_ratio/(1+P_to_N_size_ratio))*diffusion_cap;
    NMOS_diffusion = (1/(1+P_to_N_size_ratio))*diffusion_cap;

    leakage_power
    += calculate_leakage_power(NMOS, NMOS_diffusion, inverter_size);

    leakage_power
    += calculate_leakage_power(PMOS, PMOS_diffusion, inverter_size*P_to_N_size_ratio);

    return(leakage_power);
}

/***************************************************************************/
/* Calculate the leakge power for all basic components in the circuit*/
void calculate_individual_leakage(
				  int num_switch )
{
    int switch_count;
    float var;
    /*1X PMOS transistor */
    leakage_1X_nmos
    = calculate_leakage_power(NMOS, NMOS_C_drain, 1);

    /*1X NMOS transistor */
    leakage_1X_pmos
    = calculate_leakage_power(PMOS, PMOS_C_drain, 1);

    /* inverters */
    leakage_1X_inv = calculate_INV_leakage(1,  INV1x_Cout);
    leakage_2X_inv = calculate_INV_leakage(2,  INV2x_Cout);
    leakage_4x_inv = calculate_INV_leakage(4,  INV4x_Cout);
    
    //Satish. I have included SRAM cell leakage as the leakage of 2 minimum width NMOS transistors and 2 minimum width inverters. 
    SRAM_leakage=2*leakage_1X_nmos + 2*leakage_1X_inv;
    //printf("\n SRAM LEAKAGE is %g",SRAM_leakage);
    /* switches */
    for (switch_count=0; switch_count<num_switch-2; switch_count++)
      {
	switch_inf[switch_count].Leakage_Power
	  = calculate_switch_leakage ( switch_inf[switch_count].buffered,
				       switch_inf[switch_count].Cout, switch_count);

	//printf ("\n\n iSwitch: %d Leak power: %g \n",switch_count,switch_inf[switch_count].Leakage_Power);
      }
    
}//eof



/***************************************************************************/
float calculate_switch_leakage(
			       boolean buffered,
			       float Cout,
			       int iswitch)// added by cristinel.
{
    int i, total_stage;
    float fanOut;
    float stage_effort;
    float switch_leakage;
    float mid_inverter_size;
    float mid_Cout;

    switch_leakage = 0;

    /* estimate the size of the pass transitor in the switch */
    fanOut = Cout/NMOS_C_drain;

    // cristinel.ababei:  hard wired routing patterns have switches
    // with bufferes but no pass transistors. so do not add leakage
    // of the pass transistor in this case.
    // oBs:
    // *) a usual used switch-conn should leak only in buffer and sram but not 
    // pass-tran.
    // *) a usual un-used switch-conn should leak in all components.
    // *) a hwrp switch-conn (used or not) leaks only in its buffer. 

    if ( switch_inf[iswitch].is_hwrp == FALSE ) {
      switch_leakage += calculate_leakage_power(NMOS, Cout, fanOut);
    }
    //else {
      // Do nothing.
    //}
    //printf ("  switch_leakage: %g \n",switch_leakage);

    /* If the switch is a tri-state buffer, estimate the buffer
     chain using the stage effort as 4 */
    if (buffered)
    {
        total_stage = calc_num_buffer_stages(fanOut, 4.0);
        stage_effort = calc_buffer_stage_effort(total_stage, fanOut);
/*
        printf("Cap min NMOS          = %e\n", NMOS_C_drain);
        printf("Leakage: fanOut       = %f\n", fanOut);
        printf("Leakage: #stages      = %d\n", total_stage);
        printf("Leakage: stage effort = %f\n", stage_effort);
*/
        mid_inverter_size = fanOut;
        mid_Cout = (1+P_to_N_size_ratio)*Cout;

        for (i=0; i<total_stage; i++)
        {
            switch_leakage += calculate_INV_leakage(mid_inverter_size, mid_Cout);
            mid_inverter_size = mid_inverter_size/stage_effort;
            mid_Cout = mid_Cout/stage_effort;
        }
    }
    //printf ("  switch_leakage: %g \n",switch_leakage);

    return (switch_leakage);
}//eof


/***************************************************************************/

